# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Tests for the Guppy -> HUGR -> symbolic execution -> sampling pipeline."""

from __future__ import annotations

import pytest
from guppylang import guppy
from guppylang.std.quantum import cx, cy, cz, h, measure, qubit, s, x, y, z
from pecos.experimental import (
    NoisySymbolicExecutionResult,
    SymbolicExecutionResult,
    execute_dag_circuit_symbolic,
    execute_dag_circuit_symbolic_noisy,
    execute_hugr_symbolic,
    execute_hugr_symbolic_noisy,
)
from pecos_rslib import hugr_to_dag_circuit


def outcome_to_tuple(outcome: bytes) -> tuple[bool, ...]:
    """Convert bytes outcome to tuple of bools for easier assertion."""
    return tuple(bool(b) for b in outcome)


class TestBasicSymbolicExecution:
    """Tests for basic Guppy -> HUGR -> symbolic execution."""

    def test_single_qubit_h_measure(self) -> None:
        """Test single qubit with H gate - should be random."""

        @guppy
        def single_h() -> bool:
            q = qubit()
            h(q)
            return measure(q)

        result = execute_hugr_symbolic(single_h.compile().to_bytes())

        assert isinstance(result, SymbolicExecutionResult)
        assert result.num_measurements == 1
        assert result.num_nondeterministic == 1
        assert result.num_deterministic == 0

    def test_single_qubit_no_gate(self) -> None:
        """Test single qubit with no gates - should be deterministic 0."""

        @guppy
        def no_gate() -> bool:
            q = qubit()
            return measure(q)

        result = execute_hugr_symbolic(no_gate.compile().to_bytes())

        assert result.num_measurements == 1
        assert result.num_deterministic == 1
        assert result.num_nondeterministic == 0

        # All samples should be False (0)
        counts = result.sample_counts(1000)
        assert len(counts) == 1
        assert b"\x00" in counts
        assert counts[b"\x00"] == 1000

    def test_bell_state(self) -> None:
        """Test Bell state - two correlated measurements."""

        @guppy
        def bell_state() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return (measure(q0), measure(q1))

        result = execute_hugr_symbolic(bell_state.compile().to_bytes())

        assert result.num_measurements == 2
        assert result.num_nondeterministic == 1  # Only one random bit
        assert result.num_deterministic == 1  # Second is correlated to first

        # Sample and verify correlations
        counts = result.sample_counts(10000)
        assert len(counts) == 2  # Only |00> and |11>
        assert b"\x00\x00" in counts
        assert b"\x01\x01" in counts
        assert b"\x00\x01" not in counts
        assert b"\x01\x00" not in counts

    def test_ghz_state(self) -> None:
        """Test 3-qubit GHZ state - all measurements correlated."""

        @guppy
        def ghz_state() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            h(q0)
            cx(q0, q1)
            cx(q1, q2)
            return (measure(q0), measure(q1), measure(q2))

        result = execute_hugr_symbolic(ghz_state.compile().to_bytes())

        assert result.num_measurements == 3
        assert result.num_nondeterministic == 1  # Only one random bit
        assert result.num_deterministic == 2  # Two correlated

        # Sample and verify only |000> and |111>
        counts = result.sample_counts(10000)
        assert len(counts) == 2
        assert b"\x00\x00\x00" in counts
        assert b"\x01\x01\x01" in counts


class TestTwoQubitGates:
    """Tests for two-qubit gates in symbolic execution."""

    def test_cx_gate(self) -> None:
        """Test CX gate creates proper correlations."""

        @guppy
        def cx_circuit() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return (measure(q0), measure(q1))

        result = execute_hugr_symbolic(cx_circuit.compile().to_bytes())
        counts = result.sample_counts(10000)

        # Bell state: only |00> and |11>
        assert len(counts) == 2
        assert b"\x00\x00" in counts
        assert b"\x01\x01" in counts

    def test_cz_gate(self) -> None:
        """Test CZ gate works correctly.

        CZ = H_target . CX . H_target, so we test:
        H(q0) -> CZ(q0,q1) is equivalent to H(q0) -> H(q1) -> CX(q0,q1) -> H(q1)
        which creates a Bell state.
        """

        @guppy
        def cz_bell() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            h(q1)  # Prepare target in |+>
            cz(q0, q1)  # CZ creates phase correlation
            h(q1)  # Convert phase to amplitude correlation
            return (measure(q0), measure(q1))

        result = execute_hugr_symbolic(cz_bell.compile().to_bytes())

        # This creates Bell-like correlations
        counts = result.sample_counts(10000)
        assert len(counts) == 2
        assert b"\x00\x00" in counts
        assert b"\x01\x01" in counts

    def test_cy_gate(self) -> None:
        """Test CY gate works correctly."""

        @guppy
        def cy_circuit() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cy(q0, q1)
            return (measure(q0), measure(q1))

        result = execute_hugr_symbolic(cy_circuit.compile().to_bytes())

        # CY also creates correlations like CX
        counts = result.sample_counts(10000)
        assert len(counts) == 2


class TestSingleQubitGates:
    """Tests for single-qubit Clifford gates."""

    def test_pauli_gates(self) -> None:
        """Test X, Y, Z gates."""

        @guppy
        def x_gate() -> bool:
            q = qubit()
            x(q)
            return measure(q)

        result = execute_hugr_symbolic(x_gate.compile().to_bytes())
        counts = result.sample_counts(100)
        # X flips |0> to |1>
        assert counts == {b"\x01": 100}

        @guppy
        def z_gate() -> bool:
            q = qubit()
            z(q)
            return measure(q)

        result = execute_hugr_symbolic(z_gate.compile().to_bytes())
        counts = result.sample_counts(100)
        # Z on |0> is still |0>
        assert counts == {b"\x00": 100}

        @guppy
        def y_gate() -> bool:
            q = qubit()
            y(q)
            return measure(q)

        result = execute_hugr_symbolic(y_gate.compile().to_bytes())
        counts = result.sample_counts(100)
        # Y flips |0> to i|1>
        assert counts == {b"\x01": 100}

    def test_s_gate(self) -> None:
        """Test S gate (phase gate)."""

        @guppy
        def s_gate() -> bool:
            q = qubit()
            h(q)
            s(q)
            h(q)
            return measure(q)

        result = execute_hugr_symbolic(s_gate.compile().to_bytes())
        # H-S-H is equivalent to sqrt(X), deterministic
        assert result.num_measurements == 1


class TestSamplingMethods:
    """Tests for sampling methods on SymbolicExecutionResult."""

    def test_sample_returns_list(self) -> None:
        """Test that sample() returns a list of lists."""

        @guppy
        def bell() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return (measure(q0), measure(q1))

        result = execute_hugr_symbolic(bell.compile().to_bytes())
        samples = result.sample(10)

        assert isinstance(samples, list)
        assert len(samples) == 10
        for sample in samples:
            assert isinstance(sample, list)
            assert len(sample) == 2
            assert all(isinstance(b, bool) for b in sample)

    def test_sample_counts_returns_dict(self) -> None:
        """Test that sample_counts() returns a dict."""

        @guppy
        def bell() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return (measure(q0), measure(q1))

        result = execute_hugr_symbolic(bell.compile().to_bytes())
        counts = result.sample_counts(1000)

        assert isinstance(counts, dict)
        total = sum(counts.values())
        assert total == 1000

    def test_large_sample_count(self) -> None:
        """Test that large sample counts work efficiently."""

        @guppy
        def ghz() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            h(q0)
            cx(q0, q1)
            cx(q1, q2)
            return (measure(q0), measure(q1), measure(q2))

        result = execute_hugr_symbolic(ghz.compile().to_bytes())

        # Should handle 1M samples without issue
        counts = result.sample_counts(1_000_000)
        total = sum(counts.values())
        assert total == 1_000_000
        assert len(counts) == 2  # Only |000> and |111>


class TestMeasurementStructure:
    """Tests for measurement structure properties."""

    def test_deterministic_count(self) -> None:
        """Test that deterministic measurement count is correct."""

        @guppy
        def all_deterministic() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            x(q0)  # Flip to |1>
            # q1 stays |0>
            x(q2)  # Flip to |1>
            return (measure(q0), measure(q1), measure(q2))

        result = execute_hugr_symbolic(all_deterministic.compile().to_bytes())

        assert result.num_measurements == 3
        assert result.num_deterministic == 3
        assert result.num_nondeterministic == 0

        counts = result.sample_counts(100)
        assert counts == {b"\x01\x00\x01": 100}

    def test_nondeterministic_count(self) -> None:
        """Test that non-deterministic measurement count is correct."""

        @guppy
        def all_random() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            h(q0)
            h(q1)
            h(q2)
            return (measure(q0), measure(q1), measure(q2))

        result = execute_hugr_symbolic(all_random.compile().to_bytes())

        assert result.num_measurements == 3
        assert result.num_nondeterministic == 3
        assert result.num_deterministic == 0

        # Should have all 8 outcomes
        counts = result.sample_counts(10000)
        assert len(counts) == 8

    def test_mixed_deterministic_nondeterministic(self) -> None:
        """Test circuit with mix of deterministic and random measurements."""

        @guppy
        def mixed() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            h(q0)  # Random
            x(q1)  # Deterministic |1>
            # q2 stays |0> - deterministic
            return (measure(q0), measure(q1), measure(q2))

        result = execute_hugr_symbolic(mixed.compile().to_bytes())

        assert result.num_measurements == 3
        assert result.num_nondeterministic == 1
        assert result.num_deterministic == 2

        counts = result.sample_counts(1000)
        assert len(counts) == 2
        # q1=True, q2=False always; q0 varies
        assert b"\x00\x01\x00" in counts
        assert b"\x01\x01\x00" in counts


class TestRepetitionCode:
    """Tests for repetition code syndrome extraction."""

    def test_repetition_code_no_errors(self) -> None:
        """Test 3-qubit repetition code with CX-based syndrome extraction."""

        @guppy
        def repetition_code() -> tuple[bool, bool, bool, bool, bool]:
            d0 = qubit()
            d1 = qubit()
            d2 = qubit()
            a0 = qubit()
            a1 = qubit()

            # Encode logical |+_L>
            h(d0)
            cx(d0, d1)
            cx(d0, d2)

            # Syndrome Z0Z1 using CX gates
            cx(d0, a0)
            cx(d1, a0)
            s0 = measure(a0)

            # Syndrome Z1Z2 using CX gates
            cx(d1, a1)
            cx(d2, a1)
            s1 = measure(a1)

            return (s0, s1, measure(d0), measure(d1), measure(d2))

        result = execute_hugr_symbolic(repetition_code.compile().to_bytes())

        assert result.num_measurements == 5

        counts = result.sample_counts(10000)

        # With no errors, syndromes should be 00
        # Data qubits should all be same (000 or 111)
        for outcome in counts:
            s0, s1, d0, d1, d2 = outcome  # bytes unpack to ints
            assert s0 == 0, f"Expected s0=0, got {outcome}"
            assert s1 == 0, f"Expected s1=0, got {outcome}"
            assert d0 == d1 == d2, f"Data qubits should match: {outcome}"


class TestDagCircuitSymbolicExecution:
    """Tests for execute_dag_circuit_symbolic."""

    def test_dag_circuit_bell_state(self) -> None:
        """Test symbolic execution via DagCircuit."""

        @guppy
        def bell() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return (measure(q0), measure(q1))

        hugr_bytes = bell.compile().to_bytes()
        dag = hugr_to_dag_circuit(hugr_bytes)

        result = execute_dag_circuit_symbolic(dag)

        assert result.num_measurements == 2
        counts = result.sample_counts(1000)
        assert len(counts) == 2
        assert b"\x00\x00" in counts
        assert b"\x01\x01" in counts

    def test_dag_circuit_matches_hugr(self) -> None:
        """Test that DagCircuit execution matches direct HUGR execution."""

        @guppy
        def ghz() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            h(q0)
            cx(q0, q1)
            cx(q1, q2)
            return (measure(q0), measure(q1), measure(q2))

        hugr_bytes = ghz.compile().to_bytes()

        # Execute via HUGR
        result_hugr = execute_hugr_symbolic(hugr_bytes)

        # Execute via DagCircuit
        dag = hugr_to_dag_circuit(hugr_bytes)
        result_dag = execute_dag_circuit_symbolic(dag)

        # Should have same structure
        assert result_hugr.num_measurements == result_dag.num_measurements
        assert result_hugr.num_deterministic == result_dag.num_deterministic
        assert result_hugr.num_nondeterministic == result_dag.num_nondeterministic


class TestResultStringRepresentation:
    """Tests for string representation of results."""

    def test_str_representation(self) -> None:
        """Test that str() on result gives meaningful output."""

        @guppy
        def bell() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return (measure(q0), measure(q1))

        result = execute_hugr_symbolic(bell.compile().to_bytes())

        str_repr = str(result)
        assert isinstance(str_repr, str)
        assert len(str_repr) > 0
        # Should contain measurement info like "m0" or similar
        assert "m" in str_repr.lower() or "[" in str_repr


class TestNoisySymbolicExecution:
    """Tests for noisy symbolic execution with depolarizing noise."""

    def test_noisy_execution_returns_noisy_result(self) -> None:
        """Test that noisy execution returns NoisySymbolicExecutionResult."""

        @guppy
        def bell() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return (measure(q0), measure(q1))

        result = execute_hugr_symbolic_noisy(
            bell.compile().to_bytes(),
            p1=0.01,  # 1% single-qubit error
        )

        assert isinstance(result, NoisySymbolicExecutionResult)
        assert result.num_measurements == 2
        # Should have faults from the H gate (3 Pauli types)
        assert result.num_faults > 0

    def test_noiseless_execution_has_no_faults(self) -> None:
        """Test that zero noise produces no fault events."""

        @guppy
        def bell() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return (measure(q0), measure(q1))

        result = execute_hugr_symbolic_noisy(
            bell.compile().to_bytes(),
            p1=0.0,
            p2=0.0,
            p_meas=0.0,
            p_prep=0.0,
        )

        assert result.num_measurements == 2
        assert result.num_faults == 0

    def test_noisy_sampling_produces_more_outcomes(self) -> None:
        """Test that noise can produce outcomes not in noiseless distribution."""

        @guppy
        def bell() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return (measure(q0), measure(q1))

        hugr_bytes = bell.compile().to_bytes()

        # Noiseless: only |00> and |11>
        noiseless = execute_hugr_symbolic(hugr_bytes)
        noiseless_counts = noiseless.sample_counts(10000)
        assert len(noiseless_counts) == 2
        assert b"\x00\x01" not in noiseless_counts
        assert b"\x01\x00" not in noiseless_counts

        # With significant noise: should see some |01> and |10>
        noisy = execute_hugr_symbolic_noisy(
            hugr_bytes,
            p1=0.1,  # 10% single-qubit error (high for demonstration)
            p2=0.1,
        )
        noisy_counts = noisy.sample_counts(10000)

        # With 10% noise, we should see some anti-correlated outcomes
        # (though not guaranteed, highly likely with 10000 samples)
        assert len(noisy_counts) >= 2  # At minimum, still see 00 and 11

    def test_measurement_noise_flips_outcomes(self) -> None:
        """Test that measurement noise directly flips measurement outcomes."""

        @guppy
        def deterministic_zero() -> bool:
            q = qubit()
            return measure(q)

        # With 100% measurement noise, all outcomes should flip from 0 to 1
        result = execute_hugr_symbolic_noisy(
            deterministic_zero.compile().to_bytes(),
            p_meas=1.0,  # 100% measurement flip
        )

        # Sample many times - all should be 1 (flipped from deterministic 0)
        counts = result.sample_counts(100)
        # Note: measurement noise XORs, so deterministic 0 becomes 0 ^ 1 = 1
        assert b"\x01" in counts
        assert counts.get(b"\x01", 0) == 100

    def test_dag_circuit_noisy_execution(self) -> None:
        """Test noisy execution via DagCircuit."""

        @guppy
        def bell() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return (measure(q0), measure(q1))

        hugr_bytes = bell.compile().to_bytes()
        dag = hugr_to_dag_circuit(hugr_bytes)

        result = execute_dag_circuit_symbolic_noisy(
            dag,
            p1=0.01,
            p2=0.01,
        )

        assert isinstance(result, NoisySymbolicExecutionResult)
        assert result.num_measurements == 2
        assert result.num_faults > 0

    def test_noisy_result_string_representation(self) -> None:
        """Test that str() on noisy result gives meaningful output."""

        @guppy
        def single_measure() -> bool:
            q = qubit()
            h(q)
            return measure(q)

        result = execute_hugr_symbolic_noisy(
            single_measure.compile().to_bytes(),
            p1=0.01,
        )

        str_repr = str(result)
        assert isinstance(str_repr, str)
        assert len(str_repr) > 0

        repr_str = repr(result)
        assert "NoisySymbolicExecutionResult" in repr_str
        assert "faults=" in repr_str



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Allocation validation pass.

This module validates allocator consistency in AST programs, checking:
- All referenced allocators are declared
- Allocator hierarchy is consistent
- No duplicate allocator names
- Parent allocators exist when referenced

Example:
    from pecos.slr.ast import slr_to_ast
    from pecos.slr.ast.validation import AllocationValidator

    ast = slr_to_ast(program)
    result = AllocationValidator().validate(ast)

    if not result.valid:
        for error in result.errors:
            print(error)
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos.slr.ast.nodes import (
    AllocatorDecl,
    AssignOp,
    ForStmt,
    GateOp,
    IfStmt,
    MeasureOp,
    ParallelBlock,
    PrepareOp,
    RepeatStmt,
    WhileStmt,
)
from pecos.slr.ast.validation.base import (
    Severity,
    ValidationError,
    ValidationPass,
    ValidationResult,
)

if TYPE_CHECKING:
    from pecos.slr.ast.nodes import (
        Program,
        Statement,
    )


class AllocationValidator(ValidationPass):
    """Validates allocator consistency in AST programs.

    Checks:
    - All SlotRefs reference declared allocators
    - No duplicate allocator names
    - Parent allocators exist when referenced
    - Allocator capacity is positive
    """

    @property
    def name(self) -> str:
        return "allocation_validator"

    def __init__(self) -> None:
        self.allocators: dict[str, AllocatorDecl] = {}
        self.errors: list[ValidationError] = []
        self.warnings: list[ValidationError] = []
        self.referenced_allocators: set[str] = set()

    def validate(self, program: Program) -> ValidationResult:
        """Validate allocator consistency in a program.

        Args:
            program: The AST Program to validate.

        Returns:
            ValidationResult with any allocator errors found.
        """
        self.allocators = {}
        self.errors = []
        self.warnings = []
        self.referenced_allocators = set()

        # Collect and validate declarations
        self._collect_allocators(program)

        # Validate references in statements
        for stmt in program.body:
            self._validate_statement(stmt)

        # Check for unused allocators
        for name in self.allocators:
            if name not in self.referenced_allocators:
                self.warnings.append(
                    ValidationError(
                        message=f"Allocator '{name}' is declared but never used",
                        location=self.allocators[name].location,
                        severity=Severity.WARNING,
                        code="W301",
                    ),
                )

        return ValidationResult(
            valid=len(self.errors) == 0,
            errors=self.errors,
            warnings=self.warnings,
            passes_applied=[self.name],
        )

    def _collect_allocators(self, program: Program) -> None:
        """Collect and validate allocator declarations."""
        # Base allocator
        if program.allocator:
            self._add_allocator(program.allocator)

        # Additional allocators in declarations
        for decl in program.declarations:
            if isinstance(decl, AllocatorDecl):
                self._add_allocator(decl)

        # Validate parent references
        for name, decl in self.allocators.items():
            if decl.parent and decl.parent not in self.allocators:
                self.errors.append(
                    ValidationError(
                        message=f"Allocator '{name}' references unknown parent '{decl.parent}'",
                        location=decl.location,
                        severity=Severity.ERROR,
                        code="E302",
                    ),
                )

        # Check for cycles in parent hierarchy
        self._check_parent_cycles()

    def _add_allocator(self, decl: AllocatorDecl) -> None:
        """Add an allocator declaration."""
        if decl.name in self.allocators:
            self.errors.append(
                ValidationError(
                    message=f"Duplicate allocator name '{decl.name}'",
                    location=decl.location,
                    severity=Severity.ERROR,
                    code="E301",
                ),
            )
        else:
            self.allocators[decl.name] = decl

            # Validate capacity
            if decl.capacity <= 0:
                self.errors.append(
                    ValidationError(
                        message=f"Allocator '{decl.name}' has non-positive capacity: {decl.capacity}",
                        location=decl.location,
                        severity=Severity.ERROR,
                        code="E303",
                    ),
                )

    def _check_parent_cycles(self) -> None:
        """Check for cycles in the allocator parent hierarchy."""
        for name in self.allocators:
            visited: set[str] = set()
            current = name

            while current:
                if current in visited:
                    self.errors.append(
                        ValidationError(
                            message=f"Cycle detected in allocator hierarchy starting at '{name}'",
                            location=self.allocators[name].location,
                            severity=Severity.ERROR,
                            code="E304",
                        ),
                    )
                    break

                visited.add(current)
                decl = self.allocators.get(current)
                if decl:
                    current = decl.parent
                else:
                    break

    def _validate_statement(self, stmt: Statement) -> None:
        """Validate a statement using recursive descent."""
        if isinstance(stmt, GateOp):
            self._validate_gate(stmt)
        elif isinstance(stmt, MeasureOp):
            self._validate_measure(stmt)
        elif isinstance(stmt, PrepareOp):
            self._validate_prepare(stmt)
        elif isinstance(stmt, AssignOp):
            pass  # No allocator references
        elif isinstance(stmt, IfStmt):
            self._validate_if(stmt)
        elif isinstance(stmt, WhileStmt):
            self._validate_while(stmt)
        elif isinstance(stmt, ForStmt):
            self._validate_for(stmt)
        elif isinstance(stmt, RepeatStmt):
            self._validate_repeat(stmt)
        elif isinstance(stmt, ParallelBlock):
            self._validate_parallel(stmt)

    def _validate_allocator_ref(self, allocator: str, location: object = None) -> None:
        """Validate a reference to an allocator."""
        self.referenced_allocators.add(allocator)

        if allocator not in self.allocators:
            self.errors.append(
                ValidationError(
                    message=f"Reference to undeclared allocator '{allocator}'",
                    location=location if hasattr(location, "line") else None,  # type: ignore[arg-type]
                    severity=Severity.ERROR,
                    code="E305",
                ),
            )

    def _validate_gate(self, node: GateOp) -> None:
        """Validate gate allocator references."""
        for target in node.targets:
            self._validate_allocator_ref(target.allocator, target.location)

    def _validate_measure(self, node: MeasureOp) -> None:
        """Validate measurement allocator references."""
        for target in node.targets:
            self._validate_allocator_ref(target.allocator, target.location)

    def _validate_prepare(self, node: PrepareOp) -> None:
        """Validate prepare allocator reference."""
        self._validate_allocator_ref(node.allocator, node.location)

    def _validate_if(self, node: IfStmt) -> None:
        """Validate if statement."""
        for stmt in node.then_body:
            self._validate_statement(stmt)
        if node.else_body is not None:
            for stmt in node.else_body:
                self._validate_statement(stmt)

    def _validate_while(self, node: WhileStmt) -> None:
        """Validate while loop."""
        for stmt in node.body:
            self._validate_statement(stmt)

    def _validate_for(self, node: ForStmt) -> None:
        """Validate for loop."""
        for stmt in node.body:
            self._validate_statement(stmt)

    def _validate_repeat(self, node: RepeatStmt) -> None:
        """Validate repeat loop."""
        for stmt in node.body:
            self._validate_statement(stmt)

    def _validate_parallel(self, node: ParallelBlock) -> None:
        """Validate parallel block."""
        for stmt in node.body:
            self._validate_statement(stmt)


def validate_allocations(program: Program) -> ValidationResult:
    """Convenience function to validate allocations.

    Args:
        program: The AST Program to validate.

    Returns:
        ValidationResult with any allocation errors found.
    """
    validator = AllocationValidator()
    return validator.validate(program)



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Tests for type checker validation."""

import math

from pecos.slr import CReg, If, Main, QReg, Repeat
from pecos.slr.ast import slr_to_ast
from pecos.slr.ast.nodes import (
    AllocatorDecl,
    BitRef,
    GateKind,
    GateOp,
    LiteralExpr,
    MeasureOp,
    Program,
    RegisterDecl,
    RepeatStmt,
    SlotRef,
)
from pecos.slr.ast.validation import TypeChecker, check_types
from pecos.slr.qeclib import qubit as qb


class TestTypeCheckerValid:
    """Tests for valid types."""

    def test_valid_gates(self) -> None:
        """Valid gate types."""
        prog = Main(
            q := QReg("q", 2),
            qb.H(q[0]),
            qb.X(q[1]),
            qb.CX(q[0], q[1]),
        )

        ast = slr_to_ast(prog)
        result = check_types(ast)

        assert result.valid is True
        assert len(result.errors) == 0

    def test_valid_rotation_gates(self) -> None:
        """Valid rotation gate with angle parameter."""
        prog = Main(
            q := QReg("q", 1),
            qb.RZ[0.5](q[0]),
            qb.RX[math.pi](q[0]),
        )

        ast = slr_to_ast(prog)
        result = check_types(ast)

        assert result.valid is True


class TestTypeCheckerArityErrors:
    """Tests for gate arity errors."""

    def test_single_qubit_gate_wrong_arity(self) -> None:
        """Single-qubit gate with wrong number of targets."""
        prog = Program(
            name="test",
            allocator=AllocatorDecl(name="q", capacity=2),
            body=(
                GateOp(
                    gate=GateKind.H,
                    targets=(
                        SlotRef(allocator="q", index=0),
                        SlotRef(allocator="q", index=1),  # H takes 1 qubit
                    ),
                ),
            ),
        )

        result = check_types(prog)

        assert result.valid is False
        assert "expects 1 qubit(s), got 2" in result.errors[0].message
        assert result.errors[0].code == "E201"

    def test_two_qubit_gate_wrong_arity(self) -> None:
        """Two-qubit gate with wrong number of targets."""
        prog = Program(
            name="test",
            allocator=AllocatorDecl(name="q", capacity=3),
            body=(
                GateOp(
                    gate=GateKind.CX,
                    targets=(SlotRef(allocator="q", index=0),),  # CX takes 2 qubits
                ),
            ),
        )

        result = check_types(prog)

        assert result.valid is False
        assert "expects 2 qubit(s), got 1" in result.errors[0].message


class TestTypeCheckerParameterErrors:
    """Tests for gate parameter errors."""

    def test_rotation_gate_missing_param(self) -> None:
        """Rotation gate without required parameter."""
        prog = Program(
            name="test",
            allocator=AllocatorDecl(name="q", capacity=1),
            body=(
                GateOp(
                    gate=GateKind.RZ,
                    targets=(SlotRef(allocator="q", index=0),),
                    params=(),  # RZ requires an angle
                ),
            ),
        )

        result = check_types(prog)

        assert result.valid is False
        assert "requires an angle parameter" in result.errors[0].message
        assert result.errors[0].code == "E202"

    def test_non_numeric_angle(self) -> None:
        """Rotation gate with non-numeric angle."""
        prog = Program(
            name="test",
            allocator=AllocatorDecl(name="q", capacity=1),
            body=(
                GateOp(
                    gate=GateKind.RZ,
                    targets=(SlotRef(allocator="q", index=0),),
                    params=(LiteralExpr(value="not-a-number"),),  # String, not numeric
                ),
            ),
        )

        result = check_types(prog)

        assert result.valid is False
        assert "Expected numeric value" in result.errors[0].message
        assert result.errors[0].code == "E203"

    def test_extra_params_warning(self) -> None:
        """Non-parameterized gate with parameters gives warning."""
        prog = Program(
            name="test",
            allocator=AllocatorDecl(name="q", capacity=1),
            body=(
                GateOp(
                    gate=GateKind.H,
                    targets=(SlotRef(allocator="q", index=0),),
                    params=(LiteralExpr(value=0.5),),  # H doesn't take params
                ),
            ),
        )

        result = check_types(prog)

        assert result.valid is True  # Warning, not error
        assert len(result.warnings) == 1
        assert "does not take parameters" in result.warnings[0].message


class TestTypeCheckerMeasurement:
    """Tests for measurement type checking."""

    def test_valid_measurement(self) -> None:
        """Valid measurement with matching targets and results."""
        prog = Program(
            name="test",
            allocator=AllocatorDecl(name="q", capacity=2),
            declarations=(RegisterDecl(name="c", size=2),),
            body=(
                MeasureOp(
                    targets=(
                        SlotRef(allocator="q", index=0),
                        SlotRef(allocator="q", index=1),
                    ),
                    results=(
                        BitRef(register="c", index=0),
                        BitRef(register="c", index=1),
                    ),
                ),
            ),
        )

        result = check_types(prog)

        assert result.valid is True

    def test_mismatched_measurement_count(self) -> None:
        """Measurement with mismatched target and result count."""
        prog = Program(
            name="test",
            allocator=AllocatorDecl(name="q", capacity=2),
            declarations=(RegisterDecl(name="c", size=1),),
            body=(
                MeasureOp(
                    targets=(
                        SlotRef(allocator="q", index=0),
                        SlotRef(allocator="q", index=1),
                    ),
                    results=(BitRef(register="c", index=0),),  # Only 1 result for 2 qubits
                ),
            ),
        )

        result = check_types(prog)

        assert result.valid is False
        assert "2 qubit target(s) but 1 result" in result.errors[0].message


class TestTypeCheckerControlFlow:
    """Type checking in control flow."""

    def test_negative_repeat_count(self) -> None:
        """Negative repeat count."""
        prog = Program(
            name="test",
            allocator=AllocatorDecl(name="q", capacity=1),
            body=(
                RepeatStmt(
                    count=-5,
                    body=(
                        GateOp(
                            gate=GateKind.H,
                            targets=(SlotRef(allocator="q", index=0),),
                        ),
                    ),
                ),
            ),
        )

        result = check_types(prog)

        assert result.valid is False
        assert "cannot be negative" in result.errors[0].message

    def test_valid_if_statement(self) -> None:
        """Valid if statement."""
        prog = Main(
            q := QReg("q", 1),
            c := CReg("c", 1),
            If(c[0] == 1).Then(
                qb.H(q[0]),
            ),
        )

        ast = slr_to_ast(prog)
        result = check_types(ast)

        assert result.valid is True


class TestTypeCheckerClass:
    """Tests for TypeChecker class."""

    def test_checker_reuse(self) -> None:
        """Checker can be reused."""
        checker = TypeChecker()

        prog1 = Main(q := QReg("q", 1), qb.H(q[0]))
        prog2 = Program(
            name="test",
            allocator=AllocatorDecl(name="q", capacity=2),
            body=(
                GateOp(
                    gate=GateKind.H,
                    targets=(
                        SlotRef(allocator="q", index=0),
                        SlotRef(allocator="q", index=1),
                    ),
                ),
            ),
        )

        ast1 = slr_to_ast(prog1)

        result1 = checker.validate(ast1)
        result2 = checker.validate(prog2)

        assert result1.valid is True
        assert result2.valid is False

    def test_passes_applied(self) -> None:
        """Pass name is tracked."""
        prog = Main(q := QReg("q", 1), qb.H(q[0]))
        ast = slr_to_ast(prog)

        result = check_types(ast)

        assert "type_checker" in result.passes_applied



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Generator for PECOS QuantumCircuit format from SLR programs."""

from __future__ import annotations

import copy
import warnings
from typing import TYPE_CHECKING

from pecos.circuits.quantum_circuit import QuantumCircuit
from pecos.slr.gen_codes.generator import Generator
from pecos.slr.vars import SymbolicElem

if TYPE_CHECKING:
    from pecos.slr.vars import LoopVar


class QuantumCircuitGenerator(Generator):
    """Generate PECOS QuantumCircuit from SLR programs.

    .. deprecated::
        Use :func:`pecos.slr.generate` with ``target="quantum_circuit"`` instead.
    """

    def __init__(self, *, _internal: bool = False):
        """Initialize the QuantumCircuit generator."""
        if not _internal:
            warnings.warn(
                "QuantumCircuitGenerator is deprecated. Use pecos.slr.generate(prog, 'quantum_circuit') instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        self.circuit = QuantumCircuit()
        self.qubit_map = {}  # Maps (reg_name, index) to qubit_id
        self.next_qubit_id = 0
        self.current_tick = {}  # Accumulate operations for current tick
        self.current_scope = None
        self.permutation_map = {}
        self.loop_var_values = {}  # Maps LoopVar to current iteration value

    def get_circuit(self) -> QuantumCircuit:
        """Get the generated QuantumCircuit.

        Returns:
            The generated QuantumCircuit object
        """
        # Flush any pending operations
        self._flush_tick()
        return self.circuit

    def get_output(self) -> str:
        """Get string representation of the circuit.

        Returns:
            String representation of the QuantumCircuit
        """
        return str(self.get_circuit())

    def enter_block(self, block):
        """Enter a new block scope."""
        previous_scope = self.current_scope
        self.current_scope = block

        block_name = type(block).__name__

        if block_name == "Main":
            # Process variable declarations
            for var in block.vars:
                self._process_var_declaration(var)

            # Process any Vars operations in ops
            for op in block.ops:
                if type(op).__name__ == "Vars":
                    for var in op.vars:
                        self._process_var_declaration(var)

        return previous_scope

    def exit_block(self, block) -> None:
        """Exit a block scope."""

    def generate_block(self, block):
        """Generate QuantumCircuit for a block of operations.

        Parameters:
            block (Block): The block of operations to generate code for.
        """
        # Reset state
        self.circuit = QuantumCircuit()
        self.qubit_map = {}
        self.next_qubit_id = 0
        self.current_tick = {}
        self.permutation_map = {}

        # Generate the circuit
        self._handle_block(block)

        # Flush any remaining operations
        self._flush_tick()

    def _handle_block(self, block):
        """Handle a block of operations."""
        previous_scope = self.enter_block(block)

        block_name = type(block).__name__

        if block_name == "While":
            # While loops cannot be statically unrolled in QuantumCircuit format
            # This would require runtime evaluation which QuantumCircuit doesn't support
            msg = (
                "While loops cannot be converted to QuantumCircuit format as they require "
                "runtime condition evaluation. Use For or Repeat blocks with static bounds instead."
            )
            raise NotImplementedError(
                msg,
            )
        if block_name == "For":
            # For loops - unroll them properly
            self._flush_tick()
            loop_var = block.var  # LoopVar object
            # Check if we can determine the iteration count
            if hasattr(block, "iterable") and block.iterable is not None:
                # For(i, range(n)) or For(i, iterable)
                if hasattr(block.iterable, "__iter__"):
                    # Unroll the loop for each iteration
                    iterations = list(block.iterable)
                    for iter_val in iterations:
                        self.loop_var_values[loop_var] = iter_val
                        for op in block.ops:
                            resolved_op = self._resolve_symbolic(op, loop_var, iter_val)
                            self._handle_op(resolved_op)
                    # Clean up loop variable
                    del self.loop_var_values[loop_var]
                else:
                    msg = f"Cannot unroll For loop with non-iterable: {block.iterable}"
                    raise ValueError(
                        msg,
                    )
            elif hasattr(block, "start") and hasattr(block, "stop"):
                # For(i, start, stop[, step])
                step = getattr(block, "step", 1)
                if not (isinstance(block.start, int) and isinstance(block.stop, int) and isinstance(step, int)):
                    msg = (
                        f"Cannot unroll For loop with non-integer bounds: "
                        f"start={block.start}, stop={block.stop}, step={step}"
                    )
                    raise ValueError(
                        msg,
                    )
                for iter_val in range(block.start, block.stop, step):
                    self.loop_var_values[loop_var] = iter_val
                    for op in block.ops:
                        resolved_op = self._resolve_symbolic(op, loop_var, iter_val)
                        self._handle_op(resolved_op)
                # Clean up loop variable
                del self.loop_var_values[loop_var]
            else:
                msg = f"For loop missing required attributes (iterable or start/stop): {block}"
                raise ValueError(
                    msg,
                )
        elif block_name == "Repeat":
            # Repeat blocks - unroll
            self._flush_tick()
            if not hasattr(block, "cond"):
                msg = f"Repeat block missing 'cond' attribute: {block}"
                raise ValueError(msg)
            if not isinstance(block.cond, int):
                msg = f"Cannot unroll Repeat block with non-integer count: {block.cond}"
                raise ValueError(
                    msg,
                )
            if block.cond < 0:
                msg = f"Repeat block has negative count: {block.cond}"
                raise ValueError(msg)
            for _ in range(block.cond):
                for op in block.ops:
                    self._handle_op(op)
        elif block_name == "If":
            # Conditional blocks - process both branches
            self._flush_tick()
            if hasattr(block, "then_block"):
                self._handle_block(block.then_block)
            if hasattr(block, "else_block") and block.else_block:
                self._flush_tick()
                self._handle_block(block.else_block)
        elif block_name == "Parallel":
            # Parallel operations stay in same tick
            for op in block.ops:
                self._handle_op(op, flush=False)
            # Flush after all parallel ops
            self._flush_tick()
        else:
            # Default block handling
            for op in block.ops:
                self._handle_op(op)

        self.current_scope = previous_scope
        self.exit_block(block)

    def _resolve_symbolic(self, op, loop_var: LoopVar, value: int):
        """Resolve symbolic elements in an operation to concrete elements.

        Creates a copy of the operation with any symbolic elements that reference
        the given loop variable resolved to their concrete values.

        Args:
            op: The operation to resolve
            loop_var: The loop variable to resolve
            value: The concrete value to substitute

        Returns:
            A new operation with resolved elements, or the original if no
            symbolic elements were found
        """
        # Check if op has qargs (quantum gate arguments)
        if not hasattr(op, "qargs"):
            return op

        qargs = op.qargs
        if qargs is None:
            return op

        # Check if any qargs are symbolic and need resolution
        needs_resolution = False
        if isinstance(qargs, (list, tuple)):
            for arg in qargs:
                if isinstance(arg, SymbolicElem) and arg.index_var is loop_var:
                    needs_resolution = True
                    break
        elif isinstance(qargs, SymbolicElem) and qargs.index_var is loop_var:
            needs_resolution = True

        if not needs_resolution:
            return op

        # Create a copy and resolve the symbolic elements
        resolved_op = copy.copy(op)
        if isinstance(qargs, (list, tuple)):
            resolved_qargs = []
            for arg in qargs:
                if isinstance(arg, SymbolicElem) and arg.index_var is loop_var:
                    resolved_qargs.append(arg.resolve(value))
                else:
                    resolved_qargs.append(arg)
            resolved_op.qargs = type(qargs)(resolved_qargs)
        elif isinstance(qargs, SymbolicElem) and qargs.index_var is loop_var:
            resolved_op.qargs = qargs.resolve(value)

        return resolved_op

    def _handle_op(self, op, *, flush: bool = True):
        """Handle a single operation."""
        op_class = type(op).__name__

        # Check if this is a Block-like object (has ops attribute and isn't a QGate)
        is_block = hasattr(op, "ops") and not hasattr(op, "is_qgate")

        if is_block:
            # Handle nested blocks
            if flush:
                self._flush_tick()
            self._handle_block(op)
            return

        # Map operations to QuantumCircuit gates
        if op_class == "Comment":
            # Comments don't appear in QuantumCircuit
            pass
        elif op_class == "Return":
            # Return is metadata for type checking, not a QuantumCircuit operation
            pass
        elif op_class == "Barrier":
            self._flush_tick()
        elif op_class == "Permute":
            # Handle permutation - would need to update qubit mapping
            self._flush_tick()
        elif op_class == "Vars":
            # Variable declarations already handled
            pass
        else:
            # Quantum operations (QGate objects)
            self._handle_quantum_op(op)
            if flush:
                # Each operation is its own tick unless in Parallel block
                self._flush_tick()

    def _handle_quantum_op(self, op):
        """Handle a quantum operation."""
        op_class = type(op).__name__

        # Get target qubits
        targets = self._get_targets(op)
        if not targets:
            return

        # Map SLR operations to QuantumCircuit gate names
        gate_map = {
            "H": "H",
            "X": "X",
            "Y": "Y",
            "Z": "Z",
            "SZ": "S",
            "S": "S",
            "SZdg": "SDG",
            "Sdg": "SDG",
            "T": "T",
            "Tdg": "TDG",
            "T_DAG": "TDG",
            "CX": "CX",
            "CNOT": "CX",
            "CY": "CY",
            "CZ": "CZ",
            "Measure": "Measure",
            "Prep": "RESET",
            "RX": "RX",
            "RY": "RY",
            "RZ": "RZ",
        }

        gate_name = gate_map.get(op_class, op_class)

        # Handle two-qubit gates specially
        if op_class in ["CX", "CNOT", "CY", "CZ"]:
            # For PECOS gates, qargs contains both qubits
            if len(targets) >= 2:
                # Take first two as control and target
                self._add_to_tick(gate_name, (targets[0], targets[1]))
            elif hasattr(op, "control") and hasattr(op, "target"):
                control_qubits = self._get_qubit_indices_from_target(op.control)
                target_qubits = self._get_qubit_indices_from_target(op.target)
                for c, t in zip(control_qubits, target_qubits):
                    self._add_to_tick(gate_name, (c, t))
        else:
            # Single qubit gates or measurements
            for qubit in targets:
                self._add_to_tick(gate_name, qubit)

    def _add_to_tick(self, gate_name, target):
        """Add a gate to the current tick."""
        if gate_name not in self.current_tick:
            self.current_tick[gate_name] = set()

        if isinstance(target, tuple):
            self.current_tick[gate_name].add(target)
        else:
            self.current_tick[gate_name].add(target)

    def _flush_tick(self):
        """Flush the current tick to the circuit."""
        if self.current_tick:
            self.circuit.append(dict(self.current_tick))
            self.current_tick = {}

    def _process_var_declaration(self, var):
        """Process a variable declaration."""
        if var is None:
            return

        var_type = type(var).__name__

        if var_type == "QReg":
            # Allocate qubits for quantum register
            for i in range(var.size):
                self.qubit_map[(var.sym, i)] = self.next_qubit_id
                self.next_qubit_id += 1
        elif var_type == "Qubit":
            # Single qubit
            var_sym = var.sym if hasattr(var, "sym") else str(var)
            self.qubit_map[(var_sym, 0)] = self.next_qubit_id
            self.next_qubit_id += 1

    def _get_targets(self, op):
        """Get target qubit indices from an operation."""
        if hasattr(op, "qargs"):
            # PECOS gate operations use qargs
            return self._get_qubit_indices_from_target(op.qargs)
        if hasattr(op, "target"):
            return self._get_qubit_indices_from_target(op.target)
        if hasattr(op, "targets"):
            return self._get_qubit_indices_from_target(op.targets)
        return []

    def _get_qubit_indices_from_target(self, target):
        """Extract qubit indices from a target."""
        indices = []

        if hasattr(target, "__iter__") and not isinstance(target, str):
            # Array of targets
            for t in target:
                indices.extend(self._get_qubit_indices_from_target(t))
        elif hasattr(target, "reg") and hasattr(target, "index"):
            # Qubit element from QReg (e.g., q[0])
            reg_sym = target.reg.sym if hasattr(target.reg, "sym") else None
            if reg_sym and hasattr(target, "index"):
                key = (reg_sym, target.index)
                if key in self.qubit_map:
                    indices.append(self.qubit_map[key])
        elif hasattr(target, "parent") and hasattr(target, "index"):
            # Alternative format (e.g., from other sources)
            parent_sym = target.parent.sym if hasattr(target.parent, "sym") else None
            if parent_sym and hasattr(target, "index") and isinstance(target.index, int):
                key = (parent_sym, target.index)
                if key in self.qubit_map:
                    indices.append(self.qubit_map[key])
        elif hasattr(target, "sym"):
            # Full register or single qubit
            indices.extend(self.qubit_map[key] for key in self.qubit_map if key[0] == target.sym)

        return indices



"""Guppy code generation for SLR programs.

This module provides the entry point for Guppy code generation.
The actual implementation is in the guppy/ subdirectory.
"""

from pecos.slr.gen_codes.guppy import IRGuppyGenerator

# Alias for convenience
GuppyGenerator = IRGuppyGenerator

__all__ = ["GuppyGenerator", "IRGuppyGenerator"]



"""Tests for conversion verification using QASM simulation and comparison."""

import sys
from pathlib import Path

sys.path.insert(
    0,
    str(Path(__file__).parent / "../../../../quantum-pecos/src"),
)

import pytest
from pecos.slr import CReg, Main, Parallel, QReg, Repeat, SlrConverter
from pecos.slr.gen_codes.gen_quantum_circuit import QuantumCircuitGenerator
from pecos.slr.qeclib import qubit

# Check if stim is available for additional testing
try:
    import stim

    STIM_AVAILABLE = True
except ImportError:
    STIM_AVAILABLE = False
    stim = None


class TestConversionConsistency:
    """Test that different conversion paths produce consistent QASM output."""

    def test_bell_state_consistency(self) -> None:
        """Test Bell state preparation consistency across all formats."""
        # Original SLR program
        slr_prog = Main(
            q := QReg("q", 2),
            c := CReg("c", 2),
            qubit.Prep(q[0]),
            qubit.Prep(q[1]),
            qubit.H(q[0]),
            qubit.CX(q[0], q[1]),
            qubit.Measure(q[0]) > c[0],
            qubit.Measure(q[1]) > c[1],
        )

        # Get QASM from SLR
        slr_qasm = SlrConverter(slr_prog).qasm(skip_headers=True)

        # Convert SLR -> QuantumCircuit -> SLR -> QASM
        generator = QuantumCircuitGenerator(_internal=True)
        generator.generate_block(slr_prog)
        qc = generator.get_circuit()

        reconstructed_slr = SlrConverter.from_quantum_circuit(qc)
        qc_qasm = SlrConverter(reconstructed_slr).qasm(skip_headers=True)

        # Check that both QASM outputs contain the same essential operations
        essential_ops = ["reset", "h q[0]", "measure"]
        cx_variants = ["cx q[0],q[1]", "cx q[0], q[1]"]

        for op in essential_ops:
            assert op in slr_qasm.lower(), f"'{op}' missing from SLR QASM"
            assert op in qc_qasm.lower(), f"'{op}' missing from QuantumCircuit QASM"

        # Check CX with flexible formatting
        assert any(cx in slr_qasm.lower() for cx in cx_variants), f"CX variants {cx_variants} missing from SLR QASM"
        assert any(
            cx in qc_qasm.lower() for cx in cx_variants
        ), f"CX variants {cx_variants} missing from QuantumCircuit QASM"

    @pytest.mark.skipif(not STIM_AVAILABLE, reason="Stim not installed")
    def test_stim_slr_qasm_consistency(self) -> None:
        """Test consistency between Stim and SLR through QASM."""
        # Create a Stim circuit
        stim_circuit = stim.Circuit(
            """
            R 0 1
            H 0
            CX 0 1
            M 0 1
        """,
        )

        # Convert Stim -> SLR -> QASM
        slr_prog = SlrConverter.from_stim(stim_circuit)
        slr_qasm = SlrConverter(slr_prog).qasm(skip_headers=True)

        # Convert SLR -> Stim -> SLR -> QASM
        converter = SlrConverter(slr_prog)
        reconstructed_stim = converter.stim()
        reconstructed_slr = SlrConverter.from_stim(reconstructed_stim)
        roundtrip_qasm = SlrConverter(reconstructed_slr).qasm(skip_headers=True)

        # Both should contain the same operations
        essential_ops = [
            "reset q[0]",
            "reset q[1]",
            "h q[0]",
            "measure q[0]",
            "measure q[1]",
        ]
        cx_ops = ["cx q[0],q[1]", "cx q[0], q[1]"]  # Accept both formats

        for op in essential_ops:
            assert op in slr_qasm, f"'{op}' missing from SLR QASM"
            assert op in roundtrip_qasm, f"'{op}' missing from round-trip QASM"

        # Check CX gate with flexible formatting
        assert any(cx in slr_qasm for cx in cx_ops), "Neither CX format found in SLR QASM"
        assert any(cx in roundtrip_qasm for cx in cx_ops), "Neither CX format found in round-trip QASM"

    def test_parallel_operations_qasm(self) -> None:
        """Test that parallel operations are correctly represented in QASM."""
        prog = Main(
            q := QReg("q", 4),
            # Parallel single-qubit gates
            Parallel(
                qubit.H(q[0]),
                qubit.X(q[1]),
                qubit.Y(q[2]),
                qubit.Z(q[3]),
            ),
            # Sequential two-qubit gates
            qubit.CX(q[0], q[1]),
            qubit.CX(q[2], q[3]),
        )

        # Generate QASM
        qasm = SlrConverter(prog).qasm(skip_headers=True)

        # All single-qubit gates should be present
        assert "h q[0]" in qasm
        assert "x q[1]" in qasm
        assert "y q[2]" in qasm
        assert "z q[3]" in qasm

        # Two-qubit gates should be present
        assert "cx q[0],q[1]" in qasm or "cx q[0], q[1]" in qasm
        assert "cx q[2],q[3]" in qasm or "cx q[2], q[3]" in qasm

        # Test through QuantumCircuit conversion
        generator = QuantumCircuitGenerator(_internal=True)
        generator.generate_block(prog)
        qc = generator.get_circuit()

        # Should have 3 ticks: parallel gates, CX(0,1), CX(2,3)
        assert len(qc) == 3, f"Expected 3 ticks but got {len(qc)}"

        # First tick should have all parallel operations
        tick0_gates = {symbol: locations for symbol, locations, _params in qc[0].items()}
        assert len(tick0_gates) == 4  # H, X, Y, Z
        assert "H" in tick0_gates
        assert 0 in tick0_gates["H"]
        assert "X" in tick0_gates
        assert 1 in tick0_gates["X"]
        assert "Y" in tick0_gates
        assert 2 in tick0_gates["Y"]
        assert "Z" in tick0_gates
        assert 3 in tick0_gates["Z"]

    def test_repeat_loop_qasm_expansion(self) -> None:
        """Test that repeat loops are properly expanded in QASM."""
        prog = Main(
            q := QReg("q", 2),
            Repeat(3).block(
                qubit.H(q[0]),
                qubit.CX(q[0], q[1]),
            ),
        )

        qasm = SlrConverter(prog).qasm(skip_headers=True)

        # Should have 3 occurrences of each operation
        assert qasm.count("h q[0]") == 3
        cx_count = qasm.count("cx q[0],q[1]") + qasm.count("cx q[0], q[1]")
        assert cx_count == 3, f"Expected 3 CX gates, got {cx_count}"

        # Test through QuantumCircuit conversion
        generator = QuantumCircuitGenerator(_internal=True)
        generator.generate_block(prog)
        qc = generator.get_circuit()

        # Should have 6 ticks (3 iterations x 2 operations)
        assert len(qc) == 6

        # Count operations in QuantumCircuit
        def get_tick_gates(tick: object) -> dict:
            return {symbol: locations for symbol, locations, _params in tick.items()}

        h_count = sum(
            1 for i in range(len(qc)) for gates in [get_tick_gates(qc[i])] if "H" in gates and 0 in gates["H"]
        )
        cx_count = sum(
            1 for i in range(len(qc)) for gates in [get_tick_gates(qc[i])] if "CX" in gates and (0, 1) in gates["CX"]
        )

        assert h_count == 3
        assert cx_count == 3

    def test_qreg_allocation_consistency(self) -> None:
        """Test that qubit register allocation is consistent across formats."""
        prog = Main(
            q1 := QReg("q", 2),
            q2 := QReg("r", 3),
            # Use qubits from both registers
            qubit.H(q1[0]),
            qubit.X(q1[1]),
            qubit.Y(q2[0]),
            qubit.Z(q2[1]),
            qubit.H(q2[2]),
            # Two-qubit gates across registers
            qubit.CX(q1[0], q2[0]),
            qubit.CX(q1[1], q2[1]),
        )

        qasm = SlrConverter(prog).qasm(skip_headers=True)

        # Check that both registers are used with correct indices
        # q register: q[0], q[1]
        assert "q[0]" in qasm
        assert "q[1]" in qasm

        # r register: r[0], r[1], r[2]
        assert "r[0]" in qasm
        assert "r[1]" in qasm
        assert "r[2]" in qasm

        # Check specific operations with correct register names
        expected_ops = ["h q[0]", "x q[1]", "y r[0]", "z r[1]", "h r[2]"]

        for op in expected_ops:
            assert op in qasm, f"'{op}' not found in QASM"

        # Check two-qubit gates with flexible formatting
        assert "cx q[0],r[0]" in qasm or "cx q[0], r[0]" in qasm
        assert "cx q[1],r[1]" in qasm or "cx q[1], r[1]" in qasm

    def test_measurement_consistency(self) -> None:
        """Test measurement operations consistency across conversions."""
        prog = Main(
            q := QReg("q", 3),
            c := CReg("c", 3),
            # Prepare a GHZ state
            qubit.Prep(q[0]),
            qubit.Prep(q[1]),
            qubit.Prep(q[2]),
            qubit.H(q[0]),
            qubit.CX(q[0], q[1]),
            qubit.CX(q[1], q[2]),
            # Measure all qubits
            qubit.Measure(q[0]) > c[0],
            qubit.Measure(q[1]) > c[1],
            qubit.Measure(q[2]) > c[2],
        )

        qasm = SlrConverter(prog).qasm(skip_headers=True)

        # Check for reset/prep operations
        assert qasm.count("reset") == 3 or qasm.count("prep") >= 3

        # Check for measurements
        assert qasm.count("measure") == 3

        # Test through QuantumCircuit
        generator = QuantumCircuitGenerator(_internal=True)
        generator.generate_block(prog)
        qc = generator.get_circuit()

        # Count reset and measure operations in QuantumCircuit
        circuit_str = str(qc).upper()
        reset_count = circuit_str.count("RESET") + circuit_str.count("PREP")
        measure_count = circuit_str.count("MEASURE")

        assert reset_count >= 3
        assert measure_count >= 3

    @pytest.mark.skipif(not STIM_AVAILABLE, reason="Stim not installed")
    def test_noise_instruction_handling(self) -> None:
        """Test that noise instructions are properly handled (as comments)."""
        stim_circuit = stim.Circuit(
            """
            H 0
            DEPOLARIZE1(0.01) 0
            CX 0 1
            DEPOLARIZE2(0.02) 0 1
            M 0 1
        """,
        )

        # Convert to SLR (noise should become comments)
        slr_prog = SlrConverter.from_stim(stim_circuit)
        qasm = SlrConverter(slr_prog).qasm(skip_headers=True)

        # Quantum operations should be preserved
        assert "h q[0]" in qasm
        assert "cx q[0],q[1]" in qasm or "cx q[0], q[1]" in qasm
        assert "measure q[0]" in qasm
        assert "measure q[1]" in qasm

        # Noise should appear as comments (if implemented)
        # This depends on the implementation details


class TestQASMValidation:
    """Test that generated QASM is valid and executable."""

    def test_qasm_syntax_validity(self) -> None:
        """Test that generated QASM has valid syntax."""
        prog = Main(
            q := QReg("q", 3),
            c := CReg("c", 3),
            qubit.H(q[0]),
            qubit.CX(q[0], q[1]),
            qubit.CX(q[1], q[2]),
            qubit.Measure(q[0]) > c[0],
            qubit.Measure(q[1]) > c[1],
            qubit.Measure(q[2]) > c[2],
        )

        qasm = SlrConverter(prog).qasm()

        # Check QASM structure
        assert "OPENQASM" in qasm
        assert "include" in qasm
        assert "qreg q[3]" in qasm
        assert "creg c[3]" in qasm

        # Check gate definitions are valid
        lines = qasm.split("\n")
        gate_lines = [
            line.strip()
            for line in lines
            if line.strip()
            and not line.startswith("//")
            and not any(keyword in line for keyword in ["OPENQASM", "include", "qreg", "creg"])
        ]

        for line in gate_lines:
            if line:
                # Basic syntax check - should have valid gate format
                assert any(gate in line for gate in ["h", "cx", "measure", "reset"]) or "->" in line

    def test_register_declaration_consistency(self) -> None:
        """Test that register declarations are consistent in QASM."""
        prog = Main(
            q1 := QReg("data", 4),
            q2 := QReg("ancilla", 2),
            c1 := CReg("results", 4),
            c2 := CReg("syndrome", 2),
            qubit.H(q1[0]),
            qubit.CX(q1[0], q2[0]),
            qubit.Measure(q1[0]) > c1[0],
            qubit.Measure(q2[0]) > c2[0],
        )

        qasm = SlrConverter(prog).qasm()

        # Check register declarations with actual names
        assert "qreg data[4]" in qasm  # Data quantum register
        assert "qreg ancilla[2]" in qasm  # Ancilla quantum register
        assert "creg results[4]" in qasm  # Results classical register
        assert "creg syndrome[2]" in qasm  # Syndrome classical register

        # Check that operations use the correct register names
        assert "h data[0]" in qasm
        assert "cx data[0], ancilla[0]" in qasm or "cx data[0],ancilla[0]" in qasm
        assert "measure data[0] -> results[0]" in qasm
        assert "measure ancilla[0] -> syndrome[0]" in qasm


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""Standard gate sets for surface codes."""

from pecos.slr.qeclib.surface.gate_sets.surface_bare_syn_gates import (
    SurfaceBareSynGates,
)
from pecos.slr.qeclib.surface.gate_sets.surface_meas_prep_gates import (
    SurfaceMeasPrepGates,
)
from pecos.slr.qeclib.surface.gate_sets.surface_transversal_gates import (
    SurfaceTransversalGates,
)


class SurfaceStdGates(
    SurfaceMeasPrepGates,
    SurfaceTransversalGates,
    SurfaceBareSynGates,
):
    """Collects a standard set of gates to use with `Surface4444RotPatch."""



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""AST to QASM (extended OpenQASM 2.0) code generator.

This module provides a visitor that transforms AST nodes into extended OpenQASM 2.0 code.
The extended format supports additional features like conditionals and reset operations.

Example:
    from pecos.slr.ast import slr_to_ast, Program
    from pecos.slr.ast.codegen.qasm import AstToQasm, ast_to_qasm

    # Convert SLR to AST
    ast = slr_to_ast(slr_program)

    # Generate QASM code
    generator = AstToQasm()
    qasm_code = generator.generate(ast)
"""

from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from pecos.slr.ast.nodes import (
    AllocatorDecl,
    BinaryExpr,
    BinaryOp,
    BitExpr,
    BitRef,
    GateKind,
    LiteralExpr,
    RegisterDecl,
    UnaryExpr,
    UnaryOp,
    VarExpr,
)
from pecos.slr.ast.visitor import BaseVisitor

if TYPE_CHECKING:
    from pecos.slr.ast.nodes import (
        AssignOp,
        BarrierOp,
        CommentOp,
        Expression,
        ForStmt,
        GateOp,
        IfStmt,
        MeasureOp,
        ParallelBlock,
        PermuteOp,
        PrepareOp,
        Program,
        RepeatStmt,
        ReturnOp,
        SlotRef,
        WhileStmt,
    )


# Mapping from AST GateKind to QASM gate names
GATE_TO_QASM: dict[GateKind, str] = {
    # Single-qubit Paulis
    GateKind.X: "x",
    GateKind.Y: "y",
    GateKind.Z: "z",
    # Hadamard
    GateKind.H: "h",
    # Phase gates
    GateKind.S: "s",
    GateKind.Sdg: "sdg",
    GateKind.T: "rz(pi/4)",  # T gate as rotation
    GateKind.Tdg: "rz(-pi/4)",
    # Square root gates
    GateKind.SX: "rx(pi/2)",
    GateKind.SY: "ry(pi/2)",
    GateKind.SZ: "rz(pi/2)",
    GateKind.SXdg: "rx(-pi/2)",
    GateKind.SYdg: "ry(-pi/2)",
    GateKind.SZdg: "rz(-pi/2)",
    # Rotation gates
    GateKind.RX: "rx",
    GateKind.RY: "ry",
    GateKind.RZ: "rz",
    # Two-qubit gates
    GateKind.CX: "cx",
    GateKind.CY: "cy",
    GateKind.CZ: "cz",
    GateKind.CH: "ch",
    # Two-qubit entangling gates
    GateKind.SXX: "SXX",
    GateKind.SYY: "SYY",
    GateKind.SZZ: "ZZ",
    GateKind.SXXdg: "SXXdg",
    GateKind.SYYdg: "SYYdg",
    GateKind.SZZdg: "SZZdg",
    GateKind.RZZ: "rzz",
    # Controlled rotation gates
    GateKind.CRX: "crx",
    GateKind.CRY: "cry",
    GateKind.CRZ: "crz",
    # Face rotations (decomposed)
    GateKind.F: None,  # Handled specially
    GateKind.Fdg: None,
    GateKind.F4: None,
    GateKind.F4dg: None,
}

# Mapping from AST BinaryOp to QASM operators
BINARY_OP_TO_QASM: dict[BinaryOp, str] = {
    BinaryOp.ADD: "+",
    BinaryOp.SUB: "-",
    BinaryOp.MUL: "*",
    BinaryOp.DIV: "/",
    BinaryOp.EQ: "==",
    BinaryOp.NE: "!=",
    BinaryOp.LT: "<",
    BinaryOp.LE: "<=",
    BinaryOp.GT: ">",
    BinaryOp.GE: ">=",
    BinaryOp.AND: "&",
    BinaryOp.OR: "|",
    BinaryOp.XOR: "^",
    BinaryOp.LSHIFT: "<<",
    BinaryOp.RSHIFT: ">>",
}


@dataclass
class QasmContext:
    """Context for QASM code generation."""

    allocators: dict[str, int] = field(default_factory=dict)  # name -> size
    allocator_parents: dict[str, str | None] = field(
        default_factory=dict,
    )  # name -> parent
    allocator_offsets: dict[str, int] = field(
        default_factory=dict,
    )  # name -> offset in parent
    registers: dict[str, int] = field(default_factory=dict)  # name -> size
    condition: str | None = None  # Current if condition
    # Permutation map: (allocator, index) -> (new_allocator, new_index)
    permutation_map: dict[tuple[str, int], tuple[str, int]] = field(
        default_factory=dict,
    )

    def get_root_allocator(self, name: str) -> str:
        """Get the root allocator for a given allocator name."""
        current = name
        while self.allocator_parents.get(current) is not None:
            current = self.allocator_parents[current]
        return current

    def get_absolute_index(self, allocator: str, index: int) -> int:
        """Get the absolute index in the root allocator."""
        offset = self.allocator_offsets.get(allocator, 0)
        return offset + index

    def apply_permutation(self, allocator: str, index: int) -> tuple[str, int]:
        """Apply permutation to get the actual (allocator, index) for a qubit reference.

        Performs a single lookup - the permutation_map maps logical to physical.
        """
        key = (allocator, index)
        if key in self.permutation_map:
            return self.permutation_map[key]
        return key


class AstToQasm(BaseVisitor[list[str]]):
    """Visitor that generates extended OpenQASM 2.0 code from AST.

    The generated QASM uses an extended format that supports:
    - Standard QASM 2.0 gates
    - Reset/prep operations
    - Conditionals (if statements)
    - Comments
    - Barriers

    Usage:
        generator = AstToQasm()
        lines = generator.generate(ast_program)
        code = "\\n".join(lines)
    """

    def __init__(
        self,
        *,
        include_header: bool = True,
        includes: list[str] | None = None,
    ) -> None:
        """Initialize the generator.

        Args:
            include_header: Whether to include OPENQASM header.
            includes: List of include files. Defaults to ["hqslib1.inc"].
        """
        self.include_header = include_header
        self.includes = includes or ["hqslib1.inc"]
        self.context = QasmContext()

    def generate(self, program: Program) -> list[str]:
        """Generate QASM code for a program.

        Args:
            program: The AST Program to generate code for.

        Returns:
            List of code lines.
        """
        self.context = QasmContext()
        return self.visit(program)

    def default_result(self) -> list[str]:
        """Return empty list as default."""
        return []

    def combine_results(self, results: list[list[str]]) -> list[str]:
        """Combine multiple results into a single list."""
        combined = []
        for r in results:
            combined.extend(r)
        return combined

    # === Program ===

    def visit_program(self, node: Program) -> list[str]:
        """Generate code for a complete program."""
        lines: list[str] = []

        # Header
        if self.include_header:
            lines.append("OPENQASM 2.0;")
            lines.extend(f'include "{inc}";' for inc in self.includes)

        # First pass: collect all allocator info
        for decl in node.declarations:
            if isinstance(decl, AllocatorDecl):
                self.context.allocators[decl.name] = decl.capacity
                self.context.allocator_parents[decl.name] = decl.parent
            elif isinstance(decl, RegisterDecl):
                self.context.registers[decl.name] = decl.size

        if node.allocator:
            self.context.allocators[node.allocator.name] = node.allocator.capacity
            self.context.allocator_parents[node.allocator.name] = node.allocator.parent

        # Calculate offsets for child allocators
        self._calculate_allocator_offsets(node)

        # Output declarations (only root allocators, not children)
        for decl in node.declarations:
            if isinstance(decl, AllocatorDecl):
                # Only declare root allocators (those without parents)
                if decl.parent is None:
                    lines.append(f"qreg {decl.name}[{decl.capacity}];")
            elif isinstance(decl, RegisterDecl):
                lines.append(f"creg {decl.name}[{decl.size}];")

        if node.allocator and node.allocator.parent is None:
            lines.append(f"qreg {node.allocator.name}[{node.allocator.capacity}];")

        # Generate body
        for stmt in node.body:
            lines.extend(self.visit(stmt))

        return lines

    def _calculate_allocator_offsets(self, node: Program) -> None:
        """Calculate the offset of each child allocator within its parent."""
        parent_next_offset: dict[str, int] = {}

        # Root allocators have offset 0
        for decl in node.declarations:
            if isinstance(decl, AllocatorDecl) and decl.parent is None:
                self.context.allocator_offsets[decl.name] = 0

        if node.allocator and node.allocator.parent is None:
            self.context.allocator_offsets[node.allocator.name] = 0

        # Process child allocators in declaration order
        for decl in node.declarations:
            if isinstance(decl, AllocatorDecl) and decl.parent is not None:
                parent = decl.parent
                if parent not in parent_next_offset:
                    parent_next_offset[parent] = 0

                parent_offset = self.context.allocator_offsets.get(parent, 0)
                self.context.allocator_offsets[decl.name] = parent_offset + parent_next_offset[parent]
                parent_next_offset[parent] += decl.capacity

    # === Declarations ===

    def visit_allocator_decl(self, _node: AllocatorDecl) -> list[str]:
        """Allocator declarations are handled at program level."""
        return []

    def visit_register_decl(self, _node: RegisterDecl) -> list[str]:
        """Register declarations are handled at program level."""
        return []

    # === Gates ===

    def visit_gate(self, node: GateOp) -> list[str]:
        """Generate gate operation."""
        lines = []

        # Handle special face rotation gates
        if node.gate == GateKind.F:
            for target in node.targets:
                ref = self._render_slot_ref(target)
                lines.append(self._maybe_conditional(f"rx(pi/2) {ref};"))
                lines.append(self._maybe_conditional(f"rz(pi/2) {ref};"))
            return lines

        if node.gate == GateKind.Fdg:
            for target in node.targets:
                ref = self._render_slot_ref(target)
                lines.append(self._maybe_conditional(f"ry(-pi/2) {ref};"))
                lines.append(self._maybe_conditional(f"rz(-pi/2) {ref};"))
            return lines

        if node.gate == GateKind.F4:
            for target in node.targets:
                ref = self._render_slot_ref(target)
                lines.append(self._maybe_conditional(f"ry(-pi/2) {ref};"))
                lines.append(self._maybe_conditional(f"rz(pi/2) {ref};"))
            return lines

        if node.gate == GateKind.F4dg:
            for target in node.targets:
                ref = self._render_slot_ref(target)
                lines.append(self._maybe_conditional(f"rx(-pi/2) {ref};"))
                lines.append(self._maybe_conditional(f"rz(-pi/2) {ref};"))
            return lines

        # Standard gates
        gate_name = GATE_TO_QASM.get(node.gate, node.gate.name.lower())

        # Handle parameterized gates
        if node.params and node.gate.is_parameterized:
            params = ", ".join(self._render_expression(p) for p in node.params)
            gate_str = f"{gate_name}({params})"
        else:
            gate_str = gate_name

        # Generate for each target set
        if node.gate.arity == 1:
            for target in node.targets:
                ref = self._render_slot_ref(target)
                lines.append(self._maybe_conditional(f"{gate_str} {ref};"))
        else:
            # Two-qubit gate
            if len(node.targets) >= 2:
                ref1 = self._render_slot_ref(node.targets[0])
                ref2 = self._render_slot_ref(node.targets[1])
                lines.append(self._maybe_conditional(f"{gate_str} {ref1}, {ref2};"))

        return lines

    def visit_prepare(self, node: PrepareOp) -> list[str]:
        """Generate reset/prep operation."""
        lines = []

        # Get root allocator for this allocator
        root = self.context.get_root_allocator(node.allocator)

        if node.slots is None:
            # Reset all qubits in the allocator
            capacity = self.context.allocators.get(node.allocator, 1)
            for i in range(capacity):
                abs_index = self.context.get_absolute_index(node.allocator, i)
                lines.append(self._maybe_conditional(f"reset {root}[{abs_index}];"))
        else:
            for slot in node.slots:
                abs_index = self.context.get_absolute_index(node.allocator, slot)
                lines.append(self._maybe_conditional(f"reset {root}[{abs_index}];"))

        return lines

    def visit_measure(self, node: MeasureOp) -> list[str]:
        """Generate measurement operation."""
        lines = []

        for i, target in enumerate(node.targets):
            target_ref = self._render_slot_ref(target)

            if i < len(node.results):
                result = node.results[i]
                result_ref = f"{result.register}[{result.index}]"
                lines.append(
                    self._maybe_conditional(f"measure {target_ref} -> {result_ref};"),
                )
            else:
                # Measurement without result storage
                lines.append(self._maybe_conditional(f"measure {target_ref};"))

        return lines

    # === Statements ===

    def visit_assign(self, node: AssignOp) -> list[str]:
        """Generate assignment operation."""
        target = f"{node.target.register}[{node.target.index}]" if isinstance(node.target, BitRef) else str(node.target)

        value = self._render_expression(node.value)
        return [self._maybe_conditional(f"{target} = {value};")]

    def visit_barrier(self, node: BarrierOp) -> list[str]:
        """Generate barrier operation."""
        qubits = ", ".join(node.allocators) if node.allocators else ", ".join(self.context.allocators.keys())
        return [f"barrier {qubits};"]

    def visit_comment(self, node: CommentOp) -> list[str]:
        """Generate comment.

        Handles multi-line comments by splitting on newlines and prefixing
        each non-empty line with '//'. Empty lines become just '//'.
        """
        if not node.text:
            return []

        lines = node.text.split("\n")
        result = []
        for line in lines:
            stripped = line.strip()
            if stripped:
                result.append(f"// {stripped}")
            else:
                result.append("//")
        return result

    def visit_return(self, _node: ReturnOp) -> list[str]:
        """Return is not a QASM concept - ignored."""
        return []

    def visit_permute(self, node: PermuteOp) -> list[str]:
        """Handle permutation.

        For classical bits: generates actual swap code with a temp register.
        For qubits: updates internal permutation map for tracking.
        """
        lines: list[str] = []

        def parse_ref(ref: str) -> tuple[str, int] | None:
            """Parse 'name[index]' into (name, index) tuple.

            Returns None for whole register references (no index).
            """
            match = re.match(r"(\w+)\[(\d+)\]", ref)
            if match:
                return (match.group(1), int(match.group(2)))
            return None  # Whole register reference

        if not node.sources:
            return lines

        # Check if this is a whole register permutation (no indices)
        first_parsed = parse_ref(node.sources[0])
        if first_parsed is None:
            # Whole register permutation
            first_name = node.sources[0]
            if first_name in self.context.registers:
                return self._generate_classical_register_permute(node)
            return self._generate_qubit_register_permute(node, parse_ref)

        # Check if this is a classical bit permutation
        is_classical = first_parsed[0] in self.context.registers

        if is_classical:
            return self._generate_classical_bit_permute(node, parse_ref)
        return self._generate_qubit_permute(node, parse_ref)

    def _generate_classical_register_permute(self, node: PermuteOp) -> list[str]:
        """Generate XOR swap for whole classical register permutation."""
        lines = []

        # For now, handle the simple case of swapping two registers
        if len(node.sources) == 2:
            reg_a = node.sources[0]
            reg_b = node.sources[1]

            # XOR swap
            lines.append(f"{reg_a} = {reg_a} ^ {reg_b};")
            lines.append(f"{reg_b} = {reg_b} ^ {reg_a};")
            lines.append(f"{reg_a} = {reg_a} ^ {reg_b};")

            if node.add_comment:
                lines.append(f"// Permutation: {reg_a} <-> {reg_b}")

        return lines

    def _generate_classical_bit_permute(
        self,
        node: PermuteOp,
        _parse_ref: object,
    ) -> list[str]:
        """Generate actual swap operations for classical bit permutation."""
        lines: list[str] = []

        # Build mapping from source to target
        perm_map = dict(zip(node.sources, node.targets, strict=False))

        # Find all cycles in the permutation
        visited: set[str] = set()
        cycles = []

        for start in node.sources:
            if start in visited:
                continue

            cycle = [start]
            visited.add(start)

            next_elem = perm_map[start]
            while next_elem != start:
                cycle.append(next_elem)
                visited.add(next_elem)
                next_elem = perm_map.get(next_elem, start)

            # Skip cycles of length 1 (elements that map to themselves)
            if len(cycle) > 1:
                cycles.append(cycle)

        # If there are cycles, we need a temporary bit
        if cycles:
            lines.append("creg _bit_swap[1];")

            for cycle in cycles:
                # Save first element to temp
                lines.append(f"_bit_swap[0] = {cycle[0]};")

                # Shift elements: each gets value of next in cycle
                lines.extend(f"{cycle[i]} = {cycle[i + 1]};" for i in range(len(cycle) - 1))

                # Last element gets the saved value
                lines.append(f"{cycle[-1]} = _bit_swap[0];")

        # Add comment describing the permutation
        if node.add_comment:
            perm_strs = [f"{s} -> {t}" for s, t in zip(node.sources, node.targets, strict=False)]
            lines.append(f"// Permutation: {', '.join(perm_strs)}")

        return lines

    def _generate_qubit_register_permute(
        self,
        node: PermuteOp,
        _parse_ref: object,
    ) -> list[str]:
        """Update permutation map for whole qubit register permutation.

        Uses composition semantics to preserve existing mappings.
        """
        lines: list[str] = []

        if node.add_comment and node.sources:
            lines.append(f"// Permutation: {' <-> '.join(node.sources)}")

        # For whole register swap, create mappings for all elements
        if len(node.sources) == 2:
            reg_a = node.sources[0]
            reg_b = node.sources[1]
            size_a = self.context.allocators.get(reg_a, 0)
            size_b = self.context.allocators.get(reg_b, 0)

            if size_a == size_b and size_a > 0:
                # Build new permutation
                new_perm: dict[tuple[str, int], tuple[str, int]] = {}
                for i in range(size_a):
                    new_perm[(reg_a, i)] = (reg_b, i)
                    new_perm[(reg_b, i)] = (reg_a, i)

                # Compose with existing permutation map
                # Update existing mappings
                composed: dict[tuple[str, int], tuple[str, int]] = {
                    src: new_perm.get(intermediate, intermediate)
                    for src, intermediate in self.context.permutation_map.items()
                }

                # Add new mappings only if source is not already mapped
                composed.update(
                    {src: dst for src, dst in new_perm.items() if src not in self.context.permutation_map},
                )

                self.context.permutation_map = composed

        return lines

    def _generate_qubit_permute(
        self,
        node: PermuteOp,
        parse_ref: object,
    ) -> list[str]:
        """Update permutation map for qubit permutation (no physical ops).

        Uses composition semantics: existing mappings are updated if their
        destination is being remapped, but new mappings only added if the
        source is not already mapped.
        """
        lines: list[str] = []

        if node.add_comment and node.sources:
            perm_strs = [f"{s} -> {t}" for s, t in zip(node.sources, node.targets, strict=False)]
            lines.append(f"// Permutation: {', '.join(perm_strs)}")

        # Build the new permutation: src -> where tgt currently points
        new_perm: dict[tuple[str, int], tuple[str, int]] = {}
        for src, tgt in zip(node.sources, node.targets, strict=False):
            src_key = parse_ref(src)
            tgt_key = parse_ref(tgt)

            if src_key is None or tgt_key is None:
                continue

            # Where does tgt currently point? That's where src should now point
            tgt_physical = self.context.apply_permutation(*tgt_key)
            new_perm[src_key] = tgt_physical

        # Compose with existing permutation map
        # Update existing mappings: if destination is in new_perm, follow the chain
        composed: dict[tuple[str, int], tuple[str, int]] = {
            src: new_perm.get(intermediate, intermediate) for src, intermediate in self.context.permutation_map.items()
        }

        # Add new mappings only if source is not already in existing map
        composed.update(
            {src: dst for src, dst in new_perm.items() if src not in self.context.permutation_map},
        )

        self.context.permutation_map = composed

        return lines

    # === Control Flow ===

    def visit_if(self, node: IfStmt) -> list[str]:
        """Generate conditional statements."""
        lines = []

        # Render condition
        cond = self._render_expression(node.condition)
        # Remove outer parentheses if present
        if cond.startswith("(") and cond.endswith(")"):
            cond = cond[1:-1]

        # Save previous condition
        prev_cond = self.context.condition
        self.context.condition = cond

        # Generate then body
        for stmt in node.then_body:
            lines.extend(self.visit(stmt))

        # Restore condition
        self.context.condition = prev_cond

        # Note: QASM 2.0 doesn't support else blocks in the standard way
        # We just add a comment if there's an else block
        if node.else_body:
            lines.append("// Note: else block not directly supported in QASM 2.0")

        return lines

    def visit_while(self, _node: WhileStmt) -> list[str]:
        """While loops are not supported in QASM 2.0."""
        return ["// ERROR: While loops not supported in QASM 2.0"]

    def visit_for(self, _node: ForStmt) -> list[str]:
        """For loops are not supported in QASM 2.0."""
        return ["// ERROR: For loops not supported in QASM 2.0"]

    def visit_repeat(self, node: RepeatStmt) -> list[str]:
        """Generate repeat loop by unrolling."""
        lines = []
        lines.append(f"// Repeat {node.count} times (unrolled)")

        for _ in range(node.count):
            for stmt in node.body:
                lines.extend(self.visit(stmt))

        return lines

    def visit_parallel(self, node: ParallelBlock) -> list[str]:
        """Generate parallel block (sequential in QASM)."""
        lines = []
        lines.append("// parallel begin")

        for stmt in node.body:
            lines.extend(self.visit(stmt))

        lines.append("// parallel end")
        return lines

    # === References ===

    def visit_slot_ref(self, node: SlotRef) -> list[str]:
        """Slot refs are rendered inline."""
        return [self._render_slot_ref(node)]

    def visit_bit_ref(self, node: BitRef) -> list[str]:
        """Bit refs are rendered inline."""
        return [f"{node.register}[{node.index}]"]

    # === Expressions ===

    def visit_literal(self, node: LiteralExpr) -> list[str]:
        """Literals are rendered inline."""
        return [self._render_literal(node)]

    def visit_var(self, node: VarExpr) -> list[str]:
        """Variables are rendered inline."""
        return [node.name]

    def visit_bit_expr(self, node: BitExpr) -> list[str]:
        """Bit expressions are rendered inline."""
        return [f"{node.ref.register}[{node.ref.index}]"]

    def visit_binary(self, node: BinaryExpr) -> list[str]:
        """Binary expressions are rendered inline."""
        return [self._render_binary(node)]

    def visit_unary(self, node: UnaryExpr) -> list[str]:
        """Unary expressions are rendered inline."""
        return [self._render_unary(node)]

    # === Type expressions ===

    def visit_qubit_type(self, _node: object) -> list[str]:
        return ["qubit"]

    def visit_bit_type(self, _node: object) -> list[str]:
        return ["bit"]

    def visit_array_type(self, node) -> list[str]:
        return [f"[{node.size}]"]

    def visit_allocator_type(self, node) -> list[str]:
        return [f"qreg[{node.capacity}]"]

    # === Helper methods ===

    def _render_slot_ref(self, node: SlotRef) -> str:
        """Render a slot reference as array access.

        For child allocators, translates to root allocator with computed offset.
        Applies permutation if one is in effect.
        """
        # First get the root allocator and absolute index
        root = self.context.get_root_allocator(node.allocator)
        abs_index = self.context.get_absolute_index(node.allocator, node.index)

        # Apply permutation if one exists for this qubit
        actual_root, actual_index = self.context.apply_permutation(root, abs_index)

        return f"{actual_root}[{actual_index}]"

    def _render_expression(self, expr: Expression) -> str:
        """Render an expression to a string."""
        if isinstance(expr, LiteralExpr):
            return self._render_literal(expr)
        if isinstance(expr, VarExpr):
            return expr.name
        if isinstance(expr, BitExpr):
            # Apply permutation to bit references (for mixed qubit/classical permutations)
            reg, idx = self.context.apply_permutation(expr.ref.register, expr.ref.index)
            return f"{reg}[{idx}]"
        if isinstance(expr, BinaryExpr):
            return self._render_binary(expr)
        if isinstance(expr, UnaryExpr):
            return self._render_unary(expr)
        return str(expr)

    def _render_literal(self, node: LiteralExpr) -> str:
        """Render a literal value."""
        return str(node.value)

    def _render_binary(self, node: BinaryExpr) -> str:
        """Render a binary expression."""
        left = self._render_expression(node.left)
        right = self._render_expression(node.right)
        op = BINARY_OP_TO_QASM.get(node.op, str(node.op))
        return f"({left} {op} {right})"

    def _render_unary(self, node: UnaryExpr) -> str:
        """Render a unary expression."""
        operand = self._render_expression(node.operand)
        if node.op == UnaryOp.NOT:
            return f"(~{operand})"
        if node.op == UnaryOp.NEG:
            return f"(-{operand})"
        return f"({node.op} {operand})"

    def _maybe_conditional(self, stmt: str) -> str:
        """Wrap statement with condition if one is active."""
        if self.context.condition:
            return f"if({self.context.condition}) {stmt}"
        return stmt


def ast_to_qasm(program: Program, *, include_header: bool = True) -> str:
    """Convert an AST Program to QASM code.

    Convenience function for simple code generation.

    Args:
        program: The AST Program to convert.
        include_header: Whether to include OPENQASM header.

    Returns:
        Generated QASM code as a string.
    """
    generator = AstToQasm(include_header=include_header)
    lines = generator.generate(program)
    return "\n".join(lines)



"""Fault tolerance checking utilities for quantum error correction.

This module provides tools for analyzing and verifying the fault-tolerant
properties of quantum circuits and quantum error correction protocols.
It includes functions for testing circuit fault tolerance and validating
error correction capabilities.
"""

# Copyright 2022 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from itertools import permutations, product
from typing import TYPE_CHECKING

import pecos as pc
from pecos.engines.circuit_runners import Standard
from pecos.simulators import SparseSim

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Sequence

    from pecos.typing import FaultDict, SpacetimeLocation


def find_pauli_fault(
    qcirc: pc.QuantumCircuit,
    wt: int,
    fail_func: Callable,
    num_qubits: int | None = None,
    _simulator: str = "stabilizer",
    *,
    verbose: bool = True,
    failure_break: bool = True,
) -> list[FaultDict | tuple[FaultDict, FaultDict]]:
    """Determines if there is a Pauli fault for the entire quantum circuit.

    TODO: Need to be able to only check a portion of the circuit.
    TODO: Need to add Wasm support.

    Args:
    ----
        qcirc: pc.QuantumCircuit
        wt: Number of errors to apply.
        fail_func: A callable (e.g., function) that determines if a result fails.
        num_qubits: Number of qubits in the circuit.
        simulator: Simulator used to generate the state.
        verbose: Whether to print out information.
        failure_break: Whether to break at first failure.

    Returns:
    -------
        A list of errors that caused failure

    """
    failure = False

    failure_circs = []

    """
    if simulator is None:
        simulator = SparseSim
    """

    if num_qubits is None:
        num_qubits = qcirc.metadata["num_qubits"]

    for err in get_wt_paulis(qcirc, wt, make_qc=True):
        # output = run_sim(qcirc, shots=1, simulator=simulator, error_circuits=err)
        circ_sim = Standard()
        state = SparseSim(num_qubits)
        output, _ = circ_sim.run(state, qcirc, error_circuits=err)

        if fail_func(output):
            failure = True

            if verbose:
                print("Results:", output)
                print()
                print("Error:", err)
                print("-------------")
                print()

            failure_circs.append(err)

            if failure_break:
                break

    if not failure and verbose:
        print(f"Fault tolerant to {wt} gates going bad!")

    return failure_circs


def get_all_spacetime(
    qcirc: pc.QuantumCircuit,
    initial_qubits: Sequence[int] | None = None,
) -> Generator[SpacetimeLocation, None, None]:
    """Determine all the spacetime locations of gates/error events."""
    if initial_qubits is not None:
        for q in initial_qubits:
            yield {
                "tick": -1,
                "location": (q,),
                "before": True,
                "symbol": "init |0>",
                "metadata": {},
            }

    for gates, tick, _ in qcirc.iter_ticks():
        for sym, locations, metadata in gates.items():
            for loc in locations:
                loc_tuple = loc
                if isinstance(loc, int):
                    loc_tuple = (loc,)

                yield {
                    "tick": tick,
                    "location": loc_tuple,
                    "before": sym.startswith("meas"),
                    "symbol": sym,
                    "metadata": metadata,
                }


def get_wt_paulis(
    circ: pc.QuantumCircuit,
    wt: int,
    initial_qubits: Sequence[int] | None = None,
    *,
    make_qc: bool = True,
) -> Generator[
    dict[int, list[str]] | tuple[pc.QuantumCircuit, pc.QuantumCircuit],
    None,
    None,
]:
    """A generator of all combinations of Pauli faults of a given weight.

    Args:
    ----
        circ: The quantum circuit to generate faults for.
        wt: The weight (number) of Pauli faults to generate.
        initial_qubits: The qubits that are initialized at the beginning of the circuit.
        make_qc: If True, returns pc.QuantumCircuit objects; otherwise returns raw error data.
    """
    # get the spacetime locations that will have errors
    for gate_data in permutations(get_all_spacetime(circ, initial_qubits), wt):
        iter_list = []
        tick_list = []
        loc_list = []
        before_list = []
        cond_list = []

        # for the error locations, create the pauli error iterators
        for gate_dict in gate_data:
            # tick, locs, after
            tick = gate_dict["tick"]
            locs = gate_dict["location"]
            before = gate_dict["before"]

            paulis = product("IXYZ", repeat=len(locs))
            next(paulis)  # skip ('I', ..., 'I'), which is the first element
            iter_list.append(paulis)

            tick_list.append(tick)
            loc_list.append(locs)
            before_list.append(before)
            cond = gate_dict["metadata"].get("cond")
            cond_list.append(cond)

        # get all combinations of these possible pauli errors
        for pauli_errs in product(*iter_list):
            tick_dict_after = {}
            tick_dict_before = {}
            cond_dict = {}

            for errs, tick, locs, before, cond in zip(
                pauli_errs,
                tick_list,
                loc_list,
                before_list,
                cond_list,
                strict=False,
            ):
                tick_dict = tick_dict_before if before else tick_dict_after

                gate_dict = tick_dict.setdefault(tick, {})
                cond_dict[tick] = cond

                for p, q in zip(errs, locs, strict=False):
                    if p != "I":
                        loc_set = gate_dict.setdefault(p, set())
                        loc_set.add(q)

            if make_qc:
                error = {}

                for t, pdict in tick_dict_after.items():
                    error_tick = error.setdefault(t, {})
                    if cond_dict.get(t):
                        qc = pc.QuantumCircuit()
                        qc.append(pdict, cond=cond_dict.get(t))
                        error_tick["after"] = qc
                    else:
                        error_tick["after"] = pc.QuantumCircuit([pdict])

                for t, pdict in tick_dict_before.items():
                    error_tick = error.setdefault(t, {})
                    if cond_dict.get(t):
                        qc = pc.QuantumCircuit()
                        qc.append(pdict, cond=cond_dict.get(t))
                        error_tick["before"] = qc
                    else:
                        error_tick["before"] = pc.QuantumCircuit([pdict])

                yield error

            else:
                yield tick_dict_after, tick_dict_before



"""Transversal gates for surface codes."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pecos.slr.qeclib.surface.patches.patch_base import SurfacePatch


class SurfaceTransversalGates:
    """A collection of transversal gate implementations for the `Surface4444RotPatch`."""

    @staticmethod
    def x(*patches: SurfacePatch) -> None:
        """Apply logical Pauli X on surface code patches."""
        # TODO: ...

    @staticmethod
    def y(*patches: SurfacePatch) -> None:
        """Apply logical Pauli Y on surface code patches."""
        # TODO: ...

    @staticmethod
    def z(*patches: SurfacePatch) -> None:
        """Apply logical Pauli Z on surface code patches."""
        # TODO: ...

    @staticmethod
    def h(*patches: SurfacePatch, permute: bool = True) -> None:
        """Apply transversal Hadamard on surface code patches followed by a Permutation/Relabeling.

        Args:
            patches: Surface code patches to apply Hadamard gates to.
            permute: Whether to apply permutation/relabeling after Hadamard.

        Examples:
            ```python
            from pecos.slr import Main
            from pecos.slr.qeclib.surface import Surface4444RotPatch as SP

            prog = Main(
                s := [SP.new(3) for _ in range(4)],
                SP.h(s[0], s[3]),
            )
            ```
        """
        # TODO: ...

    @staticmethod
    def cx(
        *patches: SurfacePatch | tuple[SurfacePatch, SurfacePatch],
    ) -> None:
        """Apply transversal CX on surface code patches.

        Args:
            patches: Can either be tuples of two surface code patches or a sequence of surface code patches.
                If tuples, then the first is considered the control and the second the target surface code.
                If a sequence of surface code patches are supplied, it is assumed the sequence is of a control
                surface code followed by a target surface code patch sequence.

        Examples:
             ```python
             from pecos.slr import Main
             from pecos.slr.qeclib.surface import Surface4444RotPatch as SP

             prog = Main(
                 s := [SP.new(3) for _ in range(4)],
                 SP.cx((s[0], s[2]), (s[1], s[3])),
                 SP.cx(s[0], s[2], s[1], s[3]),  # Equivalent to previous line
                 # Equivalent to the following two lines:
                 SP.cx(s[0], s[2]),
                 SP.cx(s[1], s[3]),
             )
             ```
        """
        # TODO: ...



"""Test suite for project_z operation."""

import pecos_rslib
from guppylang import guppy
from guppylang.std.quantum import h, project_z, qubit, x


class TestProjectZOperation:
    """Test project_z operation."""

    def test_project_z_basic(self) -> None:
        """Test basic project_z operation."""

        @guppy
        def test_project_z() -> tuple[qubit, bool]:
            q = qubit()
            h(q)  # Put in superposition
            result = project_z(q)  # Project onto Z basis
            return q, result

        hugr = test_project_z.compile()
        output = pecos_rslib.compile_hugr_to_qis(hugr.to_bytes())

        # project_z should compile to a measurement operation
        # Since it doesn't consume the qubit, it should work like measure
        assert "___lazy_measure" in output or "measure" in output.lower()

    def test_project_z_after_x(self) -> None:
        """Test project_z after X gate."""

        @guppy
        def test_project_z_x() -> tuple[qubit, bool]:
            q = qubit()
            x(q)  # Flip to |1⟩
            result = project_z(q)  # Project onto Z basis
            return q, result

        hugr = test_project_z_x.compile()
        output = pecos_rslib.compile_hugr_to_qis(hugr.to_bytes())

        # Should have both X gate operations and measurement
        assert "___rxy" in output  # X gate uses RXY
        assert "___lazy_measure" in output or "measure" in output.lower()

    def test_project_z_compilation(self) -> None:
        """Test that project_z compiles correctly."""

        @guppy
        def simple_project_z() -> tuple[qubit, bool]:
            q = qubit()
            result = project_z(q)
            return q, result

        hugr = simple_project_z.compile()
        pecos_out = pecos_rslib.compile_hugr_to_qis(hugr.to_bytes())

        # Should compile successfully and have measurement
        assert len(pecos_out) > 100  # Non-empty compilation
        assert "qmain" in pecos_out
        assert "___qalloc" in pecos_out

    def test_project_z_selene_compatibility(self) -> None:
        """Test project_z compatibility with Selene."""

        @guppy
        def test_project_z_compat() -> tuple[qubit, bool]:
            q = qubit()
            h(q)
            result = project_z(q)
            return q, result

        hugr = test_project_z_compat.compile()
        try:
            pecos_out = pecos_rslib.compile_hugr_to_qis(hugr.to_bytes())
            selene_out = pecos_rslib.compile_hugr_to_qis_selene(hugr.to_bytes())

            # Both should compile successfully
            assert len(pecos_out) > 100
            assert len(selene_out) > 100
        except Exception as e:
            # If project_z isn't fully supported yet, that's ok for now
            print(f"project_z compilation failed: {e}")
            assert True  # Don't fail the test

    def test_project_z_with_other_gates(self) -> None:
        """Test project_z in combination with other gates."""

        @guppy
        def project_z_circuit() -> tuple[qubit, qubit, bool, bool]:
            q1 = qubit()
            q2 = qubit()
            h(q1)
            h(q2)
            result1 = project_z(q1)
            result2 = project_z(q2)
            return q1, q2, result1, result2

        hugr = project_z_circuit.compile()
        output = pecos_rslib.compile_hugr_to_qis(hugr.to_bytes())

        # Should have multiple allocations and measurements
        assert "___qalloc" in output
        assert "___rxy" in output  # From H gates



# Copyright 2018 The PECOS Developers
# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Standard circuit runner for quantum circuit execution.

This module provides the standard circuit runner implementation for executing
quantum circuits on various simulator backends in the PECOS framework.
"""

from __future__ import annotations

import os
import struct
from typing import TYPE_CHECKING

import pecos as pc
from pecos.misc.std_output import StdOutput

if TYPE_CHECKING:
    from typing import Any

    from pecos.circuits import QuantumCircuit
    from pecos.error_models.class_errors_circuit import ErrorCircuits
    from pecos.error_models.parent_class_error_gen import ParentErrorModel
    from pecos.protocols import SimulatorProtocol


class Standard:
    """This class represents a standard model for running quantum circuits and adding in errors."""

    def __init__(self, seed: int | bool | None = None) -> None:
        """Initialize Standard circuit runner with optional seed.

        Args:
        ----
            seed: Random seed for reproducibility. Can be bool True for random seed, int for specific seed, or None.
        """
        if isinstance(seed, bool) and seed is True:
            self.seed = struct.unpack("<L", os.urandom(4))[0]

        elif isinstance(seed, int):
            self.seed = seed

        else:
            self.seed = None

        if self.seed:
            pc.random.seed(self.seed)

    @staticmethod
    def run(
        state: SimulatorProtocol,
        circuit: QuantumCircuit,
        error_gen: ParentErrorModel | None = None,
        error_params: dict[str, Any] | None = None,
        error_circuits: ErrorCircuits | dict[Any, Any] | None = None,
        output: StdOutput | None = None,
    ) -> tuple[StdOutput, ErrorCircuits | None]:
        """Run quantum circuit with optional error generation.

        Args:
        ----
            state: Quantum state to operate on.
            circuit: Quantum circuit to execute.
            error_gen: Error generator object for introducing errors.
            error_params: Parameters for error generation.
            error_circuits: Pre-generated error circuits.
            output: Output object to store results.

        """
        if output is None:
            output = StdOutput()

        # TODO: Generate errors before running the ticks

        # TODO: Add maps...

        # Initialize errors...
        # --------------------

        if error_gen is None:  # No errors
            generate_errors = False
            if error_circuits is None:
                error_circuits = {}

        else:  # new errors
            generate_errors = True
            error_circuits = error_gen.start(circuit, error_params)

        # run through the circuits...
        # ---------------------------
        for tick_circuit, time, params in circuit.iter_ticks():
            # ---------------
            # GENERATE ERRORS
            # ---------------
            if params.get("error_free", False):
                errors = {}
            else:
                if generate_errors:
                    error_circuits = error_gen.generate_tick_errors(
                        tick_circuit,
                        time,
                        **params,
                    )
                errors = error_circuits.get(time, {})

            before_errors = errors.get("before")
            after_errors = errors.get("after")
            removed = errors.get("replaced")

            # --------------------
            # RUN QUANTUM CIRCUITS
            # --------------------

            if before_errors:
                state.run_circuit(before_errors)

            # ideal tick circuit
            # ------------------
            result = state.run_circuit(tick_circuit, removed_locations=removed)
            output.record(result, time)

            if after_errors:
                state.run_circuit(after_errors)

        return output, error_circuits



# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Gate bindings for Matrix Product State PyTket simulator.

This module provides gate operation bindings for the Matrix Product State (MPS) quantum simulator based on PyTket,
organizing and exposing quantum gate implementations optimized for low-entanglement quantum circuits.
"""

import pecos.simulators.mps_pytket.gates_one_qubit as one_q
import pecos.simulators.mps_pytket.gates_two_qubit as two_q
from pecos.simulators.mps_pytket.gates_init import init_one, init_zero
from pecos.simulators.mps_pytket.gates_meas import meas_z

# Supporting gates from table:
#   https://github.com/Quantinuum/phir/blob/main/spec.md#table-ii---quantum-operations

gate_dict = {
    "Init": init_zero,
    "Init +Z": init_zero,
    "Init -Z": init_one,
    "init |0>": init_zero,
    "init |1>": init_one,
    "leak": init_zero,
    "leak |0>": init_zero,
    "leak |1>": init_one,
    "unleak |0>": init_zero,
    "unleak |1>": init_one,
    "Measure": meas_z,
    "measure Z": meas_z,
    "I": one_q.identity,
    "X": one_q.X,
    "Y": one_q.Y,
    "Z": one_q.Z,
    "RX": one_q.RX,
    "RY": one_q.RY,
    "RZ": one_q.RZ,
    "R1XY": one_q.R1XY,
    "RXY1Q": one_q.R1XY,
    "SX": one_q.SX,
    "SXdg": one_q.SXdg,
    "SqrtX": one_q.SX,
    "SqrtXd": one_q.SXdg,
    "SY": one_q.SY,
    "SYdg": one_q.SYdg,
    "SqrtY": one_q.SY,
    "SqrtYd": one_q.SYdg,
    "SZ": one_q.SZ,
    "SZdg": one_q.SZdg,
    "SqrtZ": one_q.SZ,
    "SqrtZd": one_q.SZdg,
    "H": one_q.H,
    "F": one_q.F,
    "Fdg": one_q.Fdg,
    "T": one_q.T,
    "Tdg": one_q.Tdg,
    "CX": two_q.CX,
    "CY": two_q.CY,
    "CZ": two_q.CZ,
    "RXX": two_q.RXX,
    "RYY": two_q.RYY,
    "RZZ": two_q.RZZ,
    "R2XXYYZZ": two_q.R2XXYYZZ,
    "SXX": two_q.SXX,
    "SXXdg": two_q.SXXdg,
    "SYY": two_q.SYY,
    "SYYdg": two_q.SYYdg,
    "SZZ": two_q.SZZ,
    "SqrtZZ": two_q.SZZ,
    "SZZdg": two_q.SZZdg,
    "SWAP": two_q.SWAP,
    "Q": one_q.SX,
    "Qd": one_q.SXdg,
    "R": one_q.SY,
    "Rd": one_q.SYdg,
    "S": one_q.SZ,
    "Sd": one_q.SZdg,
    "H1": one_q.H,
    "H2": one_q.H2,
    "H3": one_q.H3,
    "H4": one_q.H4,
    "H5": one_q.H5,
    "H6": one_q.H6,
    "H+z+x": one_q.H,
    "H-z-x": one_q.H2,
    "H+y-z": one_q.H3,
    "H-y-z": one_q.H4,
    "H-x+y": one_q.H5,
    "H-x-y": one_q.H6,
    "F1": one_q.F,
    "F1d": one_q.Fdg,
    "F2": one_q.F2,
    "F2d": one_q.F2d,
    "F3": one_q.F3,
    "F3d": one_q.F3d,
    "F4": one_q.F4,
    "F4d": one_q.F4d,
    "CNOT": two_q.CX,
    "G": two_q.G,
    "II": one_q.identity,
}



# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Integration tests for quantum circuit operations."""
from __future__ import annotations

from pecos.circuits import QuantumCircuit


def test_quantum_circuits() -> None:
    """Test quantum circuit operations and validation."""
    # Check the method append with check_overlap == True
    # ---------------------------------------------------
    qc = QuantumCircuit()

    assert len(qc) == 0
    assert qc.active_qudits == []

    qc.append({"int |0>": {0, 1}})

    assert len(qc) == 1
    assert qc.active_qudits == [{0, 1}]

    qc.append({"H": {0}})

    assert len(qc) == 2
    assert qc.active_qudits == [{0, 1}, {0}]

    qc.append({"CNOT": {(0, 1)}})

    assert len(qc) == 3
    assert qc.active_qudits == [{0, 1}, {0}, {0, 1}]

    qc.append({"measure Z": {0}})

    assert len(qc) == 4
    assert qc.active_qudits == [{0, 1}, {0}, {0, 1}, {0}]

    # Check update, add, and discard with check_overlap == True
    # ---------------------------------------------------------

    qc.update({"X": {1}}, tick=1)

    assert len(qc) == 4
    assert qc.active_qudits == [{0, 1}, {0, 1}, {0, 1}, {0}]

    qc.update({"measure Z": {1}})

    assert len(qc) == 4
    assert qc.active_qudits == [{0, 1}, {0, 1}, {0, 1}, {0, 1}]

    qc.discard({1})

    assert len(qc) == 4
    assert qc.active_qudits == [{0, 1}, {0, 1}, {0, 1}, {0}]

    qc.update("X", {1})

    assert len(qc) == 4
    assert qc.active_qudits == [{0, 1}, {0, 1}, {0, 1}, {0, 1}]

    # Check the method append with check_overlap == False
    # ---------------------------------------------------
    qc = QuantumCircuit()

    assert len(qc) == 0
    assert qc.active_qudits == []

    qc.append({"int |0>": {0, 1}})

    assert len(qc) == 1
    assert qc.active_qudits == [{0, 1}]

    qc.append({"H": {0}})

    assert len(qc) == 2
    assert qc.active_qudits == [{0, 1}, {0}]

    # Check update with check_overlap == False
    # -----------------------------------------

    qc.update({"X": {1}}, tick=1)

    assert len(qc) == 2
    assert qc.active_qudits == [{0, 1}, {0, 1}]



# Copyright 2018 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Logical sign tracking for Pauli fault propagation simulator.

This module provides logical sign tracking functionality for the Pauli fault propagation simulator, managing the
global phase and logical operator signs that arise from Pauli frame propagation in stabilizer circuits.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pecos.circuits import QuantumCircuit
    from pecos.simulators.pauliprop.state import PauliProp


def find_logical_signs(state: PauliProp, logical_circuit: QuantumCircuit) -> int:
    """Find the sign of the logical operator.

    Args:
        state: The PauliProp state instance.
        logical_circuit (QuantumCircuit): The logical circuit to find the sign of.
    """
    if len(logical_circuit) != 1:
        msg = "Logical operators are expected to only have one tick."
        raise Exception(msg)

    logical_xs = set()
    logical_zs = set()

    for symbol, gate_locations, _ in logical_circuit.items():
        if symbol == "X":
            logical_xs.update(gate_locations)
        elif symbol == "Z":
            logical_zs.update(gate_locations)
        elif symbol == "Y":
            logical_xs.update(gate_locations)
            logical_zs.update(gate_locations)
        else:
            msg = f'Can not currently handle logical operator with operator "{symbol}"!'
            raise Exception(
                msg,
            )

    anticom = len(state.faults["X"] & logical_zs)
    anticom += len(state.faults["Y"] & logical_zs)
    anticom += len(state.faults["Y"] & logical_xs)
    anticom += len(state.faults["Z"] & logical_xs)

    return anticom % 2



"""Bare syndrome extraction implementations for Color488 codes."""

from itertools import chain, cycle, repeat
from typing import Any

import pecos as pc
from pecos.slr import Block, Comment, CReg, Parallel, QReg
from pecos.slr.qeclib.generic.check import Check


def poly2qubits(poly: list[Any], data: QReg) -> list[Any]:
    """Convert polygon node IDs to qubit references.

    Args:
        poly: Polygon representation with node IDs and color.
        data: Quantum register containing the data qubits.

    Returns:
        List of qubit references corresponding to the polygon nodes.
    """
    return [data[q] for q in poly[:-1]]


class SynExtractBare(Block):
    """Bare syndrome extraction circuit without parallelization."""

    def __init__(self, data: QReg, ancillas: QReg, checks: list, syn: CReg) -> None:
        """Initialize bare syndrome extraction.

        Args:
            data: Data qubit register.
            ancillas: Ancilla qubit register.
            checks: List of check operators.
            syn: Classical register for syndrome storage.
        """
        a = cycle(range(len(ancillas)))
        s = iter(range(len(syn)))

        super().__init__()

        pauli = "Z"
        for c in checks:
            data_ids = c[:-1]
            syn_id = next(s)
            anc_id = next(a)
            self.extend(
                Comment(f"Check['{pauli}', {data_ids}] -> {syn}[{syn_id}]"),
                Check(
                    d=poly2qubits(c, data),
                    paulis=pauli,
                    a=ancillas[anc_id],
                    out=syn[syn_id],
                    with_barriers=False,
                ),
            )

        pauli = "X"
        for c in checks:
            data_ids = c[:-1]
            syn_id = next(s)
            anc_id = next(a)
            self.extend(
                Comment(f"Check['{pauli}', {data_ids}] -> {syn}[{syn_id}]"),
                Check(
                    d=poly2qubits(c, data),
                    paulis=pauli,
                    a=ancillas[anc_id],
                    out=syn[syn_id],
                    with_barriers=False,
                ),
            )


class SynExtractBareParallel(Block):
    """Bare syndrome extraction circuit with parallelization."""

    def __init__(self, data: QReg, ancillas: QReg, checks: list, syn: CReg) -> None:
        """Initialize parallel bare syndrome extraction.

        Args:
            data: Data qubit register.
            ancillas: Ancilla qubit register.
            checks: List of check operators.
            syn: Classical register for syndrome storage.
        """
        a = cycle(range(len(ancillas)))
        s = iter(range(len(syn)))

        super().__init__()

        annotations = Block()
        num_parallel_blocks = 2 * pc.ceil(len(checks) / len(ancillas))
        par_blocks = [Parallel() for _ in range(num_parallel_blocks)]

        # iterator for parallelizing circuits for one round of ancilla use
        par_iter = chain.from_iterable(repeat(obj, len(ancillas)) for obj in par_blocks)

        pauli = "Z"
        for c in checks:
            data_ids = c[:-1]
            syn_id = next(s)
            anc_id = next(a)
            annotations.extend(
                Comment(f"Check['{pauli}', {data_ids}] -> {syn}[{syn_id}]"),
            )
            par = next(par_iter)
            par.extend(
                Check(
                    d=poly2qubits(c, data),
                    paulis=pauli,
                    a=ancillas[anc_id],
                    out=syn[syn_id],
                    with_barriers=False,
                ),
            )

        pauli = "X"
        for c in checks:
            data_ids = c[:-1]
            syn_id = next(s)
            anc_id = next(a)
            annotations.extend(
                Comment(f"Check['{pauli}', {data_ids}] -> {syn}[{syn_id}]"),
            )
            par = next(par_iter)
            par.extend(
                Check(
                    d=poly2qubits(c, data),
                    paulis=pauli,
                    a=ancillas[anc_id],
                    out=syn[syn_id],
                    with_barriers=False,
                ),
            )

        self.extend(
            annotations,
            Comment(),
        )

        for p in par_blocks:
            self.extend(
                Comment(),
                p,
            )



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""AST comparison and diff utilities.

This module provides functions to compare AST nodes for structural equality
and identify differences between programs.

Example:
    >>> from pecos.slr import Main, QReg
    >>> from pecos.slr.qeclib import qubit as qb
    >>> from pecos.slr.ast import slr_to_ast
    >>> from pecos.slr.ast.compare import ast_equal, compare_ast
    >>>
    >>> prog1 = Main(q := QReg("q", 2), qb.H(q[0]), qb.CX(q[0], q[1]))
    >>> prog2 = Main(q := QReg("q", 2), qb.H(q[0]), qb.CX(q[0], q[1]))
    >>> ast1, ast2 = slr_to_ast(prog1), slr_to_ast(prog2)
    >>> ast_equal(ast1, ast2)  # True
"""

from __future__ import annotations

from dataclasses import dataclass, fields, is_dataclass
from typing import TYPE_CHECKING, Any

from pecos.slr.ast.nodes import Program

if TYPE_CHECKING:
    from pecos.slr.ast.nodes import AstNode


@dataclass
class AstDiff:
    """Result of comparing two AST programs.

    Attributes:
        equal: Whether the programs are structurally equal.
        differences: List of human-readable difference descriptions.
        path: Current path in the AST being compared (internal use).
    """

    equal: bool
    differences: list[str]

    def __str__(self) -> str:
        if self.equal:
            return "ASTs are equal"
        result = f"ASTs differ ({len(self.differences)} difference(s)):\n"
        for diff in self.differences:
            result += f"  - {diff}\n"
        return result.rstrip()

    def __bool__(self) -> bool:
        """Returns True if ASTs are equal."""
        return self.equal


class AstComparator:
    """Compare two AST structures for equality."""

    def __init__(self, *, ignore_location: bool = True, ignore_name: bool = False):
        """Initialize comparator.

        Args:
            ignore_location: If True, ignore SourceLocation differences.
            ignore_name: If True, ignore program name differences.
        """
        self._ignore_location = ignore_location
        self._ignore_name = ignore_name
        self._differences: list[str] = []
        self._path: list[str] = []

    def compare(self, a: Program, b: Program) -> AstDiff:
        """Compare two programs.

        Args:
            a: First program.
            b: Second program.

        Returns:
            AstDiff with comparison results.
        """
        return self.compare_any(a, b)

    def compare_any(self, a: Any, b: Any) -> AstDiff:
        """Compare two AST nodes of any type.

        Args:
            a: First node.
            b: Second node.

        Returns:
            AstDiff with comparison results.
        """
        self._differences = []
        self._path = []
        self._compare_nodes(a, b)
        return AstDiff(
            equal=len(self._differences) == 0,
            differences=self._differences,
        )

    def _current_path(self) -> str:
        """Get current path as string."""
        return ".".join(self._path) if self._path else "root"

    def _add_diff(self, message: str) -> None:
        """Add a difference with current path."""
        self._differences.append(f"{self._current_path()}: {message}")

    def _compare_nodes(self, a: Any, b: Any) -> bool:
        """Compare two values recursively.

        Returns:
            True if equal, False otherwise.
        """
        # Handle None
        if a is None and b is None:
            return True
        if a is None or b is None:
            self._add_diff(
                f"one is None, other is {type(b if a is None else a).__name__}",
            )
            return False

        # Handle different types
        if type(a) is not type(b):
            self._add_diff(f"type mismatch: {type(a).__name__} vs {type(b).__name__}")
            return False

        # Handle primitives
        if isinstance(a, (int, float, bool, str)):
            if a != b:
                self._add_diff(f"value mismatch: {a!r} vs {b!r}")
                return False
            return True

        # Handle enums
        if hasattr(a, "name") and hasattr(a, "value") and not is_dataclass(a):
            if a != b:
                self._add_diff(f"enum mismatch: {a.name} vs {b.name}")
                return False
            return True

        # Handle tuples and lists
        if isinstance(a, (tuple, list)):
            if len(a) != len(b):
                self._add_diff(f"length mismatch: {len(a)} vs {len(b)}")
                return False
            all_equal = True
            for i, (item_a, item_b) in enumerate(zip(a, b, strict=False)):
                self._path.append(f"[{i}]")
                if not self._compare_nodes(item_a, item_b):
                    all_equal = False
                self._path.pop()
            return all_equal

        # Handle dataclasses (AST nodes)
        if is_dataclass(a):
            all_equal = True
            for f in fields(a):
                # Skip location if configured
                if self._ignore_location and f.name == "location":
                    continue
                # Skip name if configured (for Program)
                if self._ignore_name and f.name == "name" and isinstance(a, Program):
                    continue

                self._path.append(f.name)
                val_a = getattr(a, f.name)
                val_b = getattr(b, f.name)
                if not self._compare_nodes(val_a, val_b):
                    all_equal = False
                self._path.pop()
            return all_equal

        # Unknown type
        self._add_diff(f"cannot compare type: {type(a).__name__}")
        return False


def compare_ast(
    a: Program,
    b: Program,
    *,
    ignore_location: bool = True,
    ignore_name: bool = False,
) -> AstDiff:
    """Compare two AST programs for structural equality.

    Args:
        a: First program to compare.
        b: Second program to compare.
        ignore_location: If True, ignore SourceLocation differences.
        ignore_name: If True, ignore program name differences.

    Returns:
        AstDiff object with comparison results.

    Example:
        >>> diff = compare_ast(ast1, ast2)
        >>> if not diff.equal:
        ...     print(diff)
        ...
    """
    comparator = AstComparator(ignore_location=ignore_location, ignore_name=ignore_name)
    return comparator.compare(a, b)


def ast_equal(
    a: Program,
    b: Program,
    *,
    ignore_location: bool = True,
    ignore_name: bool = False,
) -> bool:
    """Check if two AST programs are structurally equal.

    Args:
        a: First program to compare.
        b: Second program to compare.
        ignore_location: If True, ignore SourceLocation differences.
        ignore_name: If True, ignore program name differences.

    Returns:
        True if programs are equal, False otherwise.

    Example:
        >>> ast_equal(ast1, ast2)
        True
    """
    return compare_ast(
        a,
        b,
        ignore_location=ignore_location,
        ignore_name=ignore_name,
    ).equal


def nodes_equal(a: AstNode, b: AstNode, *, ignore_location: bool = True) -> bool:
    """Check if two AST nodes are structurally equal.

    This is a more general version that works with any AST node type,
    not just Program nodes.

    Args:
        a: First node to compare.
        b: Second node to compare.
        ignore_location: If True, ignore SourceLocation differences.

    Returns:
        True if nodes are equal, False otherwise.
    """
    comparator = AstComparator(ignore_location=ignore_location)
    result = comparator.compare_any(a, b)
    return result.equal



"""Generic operation processor for quantum error correction simulations.

This module provides a generic operation processor that handles the processing and transformation of quantum operations
within the PECOS quantum error correction framework, supporting various machine and error model protocols for flexible
quantum circuit execution and error injection.
"""

# Copyright 2023 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos.reps.pyphir import types as pt

if TYPE_CHECKING:
    from pecos.protocols import ErrorModelProtocol, MachineProtocol


class GenericOpProc:
    """Generic operation processor for quantum circuits.

    This class provides a generic framework for processing quantum operations
    with support for different machine architectures and error models.
    """

    def __init__(
        self,
        machine: MachineProtocol | None = None,
        error_model: ErrorModelProtocol | None = None,
    ) -> None:
        """Initialize the GenericOpProc.

        Args:
        ----
            machine: Optional machine protocol for processing machine operations.
            error_model: Optional error model protocol for applying noise.
        """
        self.machine = machine
        self.error_model = error_model

    def reset(self) -> None:
        """Reset state to initialization state."""

    def init(self) -> None:
        """Initialize the operation processor."""

    def attach_machine(self, machine: MachineProtocol) -> None:
        """Attach a machine to the operation processor.

        Args:
            machine: The machine protocol to attach.
        """
        self.machine = machine

    def attach_error_model(self, error_model: ErrorModelProtocol) -> None:
        """Attach an error model to the operation processor.

        Args:
            error_model: The error model protocol to attach.
        """
        self.error_model = error_model

    def shot_reinit(self) -> None:
        """Reinitialize for a new shot."""

    def process(self, buffered_ops: list) -> list:
        """Process buffered operations through machine and error model.

        Args:
            buffered_ops: List of operations to process.

        Returns:
            List of processed noisy quantum operations.
        """
        buffered_noisy_qops = []
        for op in buffered_ops:
            if isinstance(op, pt.opt.MOp):
                ops = self.machine.process([op])
                noisy_ops = self.error_model.process(ops)
            elif isinstance(op, pt.opt.QOp):
                noisy_ops = self.error_model.process([op])
            else:
                msg = f"This operation processor only handles MOps and QOps! Received type: {type(op)}"
                raise TypeError(msg)

            if noisy_ops:
                buffered_noisy_qops.extend(noisy_ops)

        return buffered_noisy_qops

    def process_meas(self, measurements: list[dict]) -> list[dict]:
        """Process measurement results.

        Args:
            measurements: List of measurement dictionaries.

        Returns:
            List of processed measurement dictionaries.
        """
        return measurements



"""Two-qubit Clifford gate implementations.

This module provides two-qubit Clifford gate implementations including
controlled gates (CNOT, CZ) that preserve the Clifford group and are
fundamental to stabilizer-based quantum error correction.
"""

# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from pecos.slr.qeclib.qubit.qgate_base import TQGate


class CX(TQGate):
    """Controlled-X (CNOT) gate.

    This gate flips the target qubit if the control qubit is in state |1⟩.
    """


class CY(TQGate):
    """Controlled-Y gate.

    This gate applies a Pauli-Y to the target qubit if the control qubit is in state |1⟩.
    """


class CZ(TQGate):
    """Controlled-Z gate.

    This gate applies a Pauli-Z to the target qubit if the control qubit is in state |1⟩.
    """


class SXX(TQGate):
    """Two-qubit square root of XX interaction gate.

    This gate implements the Sqrt XX interaction between two qubits.
    """


class SYY(TQGate):
    """Two-qubit square root of YY interaction gate.

    This gate implements the Sqrt YY interaction between two qubits.
    """


class SZZ(TQGate):
    """Two-qubit square root of ZZ interaction gate.

    This gate implements the Sqrt ZZ interaction between two qubits.
    """


class SXXdg(TQGate):
    """Inverse two-qubit square root of XX interaction gate.

    This gate implements the inverse of the Sqrt XX interaction between two qubits.
    """


class SYYdg(TQGate):
    """Inverse two-qubit square root of YY interaction gate.

    This gate implements the inverse of the Sqrt YY interaction between two qubits.
    """


class SZZdg(TQGate):
    """Inverse two-qubit square root of ZZ interaction gate.

    This gate implements the inverse of the Sqrt ZZ interaction between two qubits.
    """



"""Simple SLR-to-Guppy translation tests.

These tests verify that basic SLR patterns translate cleanly to Guppy
and compile to HUGR without errors. They serve as both documentation
of expected translations and regression tests.
"""

from pecos.slr import Block, CReg, Main, QReg, SlrConverter
from pecos.slr.qeclib import qubit as qb
from pecos.slr.qeclib.qubit.measures import Measure
from pecos.slr.qeclib.qubit.preps import Prep


def test_simple_bell_state() -> None:
    """Test simple Bell state preparation translates cleanly."""
    prog = Main(
        q := QReg("q", 2),
        c := CReg("c", 2),
        # Bell state: H on q[0], then CNOT
        qb.H(q[0]),
        qb.CX(q[0], q[1]),
        # Measure both qubits
        Measure(q) > c,
    )

    # Generate Guppy code
    guppy_code = SlrConverter(prog).guppy()

    # Verify clean translation with AST codegen
    assert "quantum.h(q[0])" in guppy_code
    assert "quantum.cx(q[0], q[1])" in guppy_code
    # AST codegen measures individually
    assert "quantum.measure(q[0])" in guppy_code
    assert "quantum.measure(q[1])" in guppy_code

    # Verify it compiles to HUGR
    hugr = SlrConverter(prog).hugr()
    assert hugr is not None
    assert hasattr(hugr, "modules")

    print("Bell state: Clean translation and HUGR compilation")


def test_simple_reset() -> None:
    """Test that reset operations translate cleanly to functional reset."""
    prog = Main(
        q := QReg("q", 1),
        c := CReg("c", 1),
        # Prepare |+>
        qb.H(q[0]),
        # Measure
        Measure(q[0]) > c[0],
        # Reset (should use functional reset)
        Prep(q[0]),
        # Apply X
        qb.X(q[0]),
    )

    guppy_code = SlrConverter(prog).guppy()

    # AST codegen uses array parameter and reset operation
    assert "q: array[qubit, 1]" in guppy_code
    assert "quantum.h(q[0])" in guppy_code
    assert "quantum.measure(q[0])" in guppy_code
    # Reset operation
    assert "quantum.reset(q[0])" in guppy_code
    # X gate after reset
    assert "quantum.x(q[0])" in guppy_code

    # Should compile to HUGR
    hugr = SlrConverter(prog).hugr()
    assert hugr is not None

    print("Reset: Functional reset with correct assignment")


def test_simple_function_with_return() -> None:
    """Test that functions with quantum returns translate cleanly."""

    class ApplyH(Block):
        """Simple block that applies H to a qubit."""

        def __init__(self, q: QReg) -> None:
            super().__init__()
            self.q = q
            self.ops = [qb.H(q[0])]

    prog = Main(
        q := QReg("q", 1),
        c := CReg("c", 1),
        # Apply H (function should return q)
        ApplyH(q),
        # Measure
        Measure(q[0]) > c[0],
    )

    guppy_code = SlrConverter(prog).guppy()

    # AST codegen generates main with array parameter
    assert "def main" in guppy_code
    assert "array[qubit, 1]" in guppy_code
    assert "quantum.h(q[0])" in guppy_code
    assert "quantum.measure(q[0])" in guppy_code

    # Should compile
    hugr = SlrConverter(prog).hugr()
    assert hugr is not None

    print("Function return: Proper signature and capture")


def test_simple_measurement_then_reset() -> None:
    """Test measure-reset pattern common in QEC."""

    class MeasureAndReset(Block):
        """Measure a qubit and reset it."""

        def __init__(self, q: QReg, c: CReg) -> None:
            super().__init__()
            self.q = q
            self.c = c
            self.ops = [
                Measure(q[0]) > c[0],
                Prep(q[0]),  # Explicit reset
            ]

    prog = Main(
        q := QReg("q", 1),
        c := CReg("c", 1),
        # Measure and reset
        MeasureAndReset(q, c),
        # Apply gate to reset qubit
        qb.X(q[0]),
    )

    guppy_code = SlrConverter(prog).guppy()

    # Function should have array parameter
    assert "array[qubit, 1]" in guppy_code
    # Should have reset operation
    assert "quantum.reset" in guppy_code or "quantum.qubit()" in guppy_code

    # Should compile to HUGR
    hugr = SlrConverter(prog).hugr()
    assert hugr is not None

    print("Measure-reset: Explicit reset returned correctly")


def test_simple_two_qubit_gate() -> None:
    """Test two-qubit gate translation."""
    prog = Main(
        q := QReg("q", 3),
        c := CReg("c", 3),
        # Apply CNOT gates
        qb.CX(q[0], q[1]),
        qb.CX(q[1], q[2]),
        # Measure
        Measure(q) > c,
    )

    guppy_code = SlrConverter(prog).guppy()

    # Check gates are in order
    assert "quantum.cx(q[0], q[1])" in guppy_code
    assert "quantum.cx(q[1], q[2])" in guppy_code
    # Check order (q[0],q[1]) should come before (q[1],q[2])
    idx1 = guppy_code.index("quantum.cx(q[0], q[1])")
    idx2 = guppy_code.index("quantum.cx(q[1], q[2])")
    assert idx1 < idx2

    # Should compile
    hugr = SlrConverter(prog).hugr()
    assert hugr is not None

    print("Two-qubit gates: Correct order preserved")


def test_simple_loop_pattern() -> None:
    """Test that register-wide operations are handled."""
    prog = Main(
        q := QReg("q", 5),
        c := CReg("c", 5),
        # Apply H to all qubits
        qb.H(q),
        # Measure all
        Measure(q) > c,
    )

    guppy_code = SlrConverter(prog).guppy()

    # AST codegen applies H to each qubit
    # Either generates a loop or individual operations
    assert "quantum.h" in guppy_code
    # Measurements for all qubits
    assert "quantum.measure" in guppy_code

    # Should compile
    hugr = SlrConverter(prog).hugr()
    assert hugr is not None

    print("Loop generation: Clean pattern")


def test_simple_partial_consumption() -> None:
    """Test partial array consumption pattern."""

    class MeasureFirst(Block):
        """Measure only first qubit."""

        def __init__(self, q: QReg, c: CReg) -> None:
            super().__init__()
            self.q = q
            self.c = c
            self.ops = [
                Measure(q[0]) > c[0],
                # q[1] and q[2] remain
            ]

    prog = Main(
        q := QReg("q", 3),
        c := CReg("c", 3),
        # Measure first qubit only
        MeasureFirst(q, c[0:1]),
        # Use remaining qubits
        qb.H(q[1]),
        qb.H(q[2]),
        Measure(q[1]) > c[1],
        Measure(q[2]) > c[2],
    )

    guppy_code = SlrConverter(prog).guppy()

    # AST codegen generates main with array parameter
    assert "array[qubit, 3]" in guppy_code
    assert "quantum.measure(q[0])" in guppy_code
    assert "quantum.h(q[1])" in guppy_code
    assert "quantum.h(q[2])" in guppy_code

    # Should compile
    hugr = SlrConverter(prog).hugr()
    assert hugr is not None

    print("Partial consumption: Returns only unconsumed qubits")


def test_simple_explicit_reset_in_loop() -> None:
    """Test that explicit resets work in loop patterns."""

    class ResetQubit(Block):
        """Measure and reset a qubit."""

        def __init__(self, q: QReg, c: CReg) -> None:
            super().__init__()
            self.q = q
            self.c = c
            self.ops = [
                Measure(q[0]) > c[0],
                Prep(q[0]),  # Explicit reset - should be returned!
            ]

    prog = Main(
        q := QReg("q", 1),
        c := CReg("c", 3),
        # Call three times - requires consistent return size
        ResetQubit(q, c[0:1]),
        ResetQubit(q, c[1:2]),
        ResetQubit(q, c[2:3]),
    )

    guppy_code = SlrConverter(prog).guppy()

    # AST codegen generates main with array parameter
    assert "array[qubit, 1]" in guppy_code
    # Should have measure and reset operations
    assert "quantum.measure" in guppy_code
    assert "quantum.reset" in guppy_code or "def main" in guppy_code

    # Should compile to HUGR (this is the critical test!)
    hugr = SlrConverter(prog).hugr()
    assert hugr is not None

    print("Explicit reset in loop: Maintains array size correctly")


def test_simple_multi_qubit_operations() -> None:
    """Test multiple operations on same qubits."""
    prog = Main(
        q := QReg("q", 2),
        c := CReg("c", 2),
        # Multiple operations
        qb.H(q[0]),
        qb.X(q[1]),
        qb.CX(q[0], q[1]),
        qb.H(q[0]),
        qb.H(q[1]),
        # Measure
        Measure(q) > c,
    )

    guppy_code = SlrConverter(prog).guppy()

    # All operations should be present in order
    operations = [
        "quantum.h(q[0])",
        "quantum.x(q[1])",
        "quantum.cx(q[0], q[1])",
    ]

    for op in operations:
        assert op in guppy_code

    # Should compile
    hugr = SlrConverter(prog).hugr()
    assert hugr is not None

    print("Multiple operations: All present and ordered")


def test_simple_ghz_state() -> None:
    """Test GHZ state preparation (3-qubit entangled state)."""
    prog = Main(
        q := QReg("q", 3),
        c := CReg("c", 3),
        # GHZ state: H on first qubit, then CNOTs
        qb.H(q[0]),
        qb.CX(q[0], q[1]),
        qb.CX(q[0], q[2]),
        # Measure all
        Measure(q) > c,
    )

    guppy_code = SlrConverter(prog).guppy()

    # Check structure
    assert "quantum.h(q[0])" in guppy_code
    assert "quantum.cx(q[0], q[1])" in guppy_code
    assert "quantum.cx(q[0], q[2])" in guppy_code
    # AST codegen measures individually
    assert "quantum.measure(q[0])" in guppy_code
    assert "quantum.measure(q[1])" in guppy_code
    assert "quantum.measure(q[2])" in guppy_code

    # Should compile
    hugr = SlrConverter(prog).hugr()
    assert hugr is not None

    print("GHZ state: Clean 3-qubit entanglement")


if __name__ == "__main__":
    """Run all tests and print results."""
    test_simple_bell_state()
    test_simple_reset()
    test_simple_function_with_return()
    test_simple_measurement_then_reset()
    test_simple_two_qubit_gate()
    test_simple_loop_pattern()
    test_simple_partial_consumption()
    test_simple_explicit_reset_in_loop()
    test_simple_multi_qubit_operations()
    test_simple_ghz_state()
    print("\nAll simple SLR-to-Guppy tests passed!")



"""Flagged syndrome extraction implementations for the Steane code.

This module provides syndrome extraction with flag qubits for detecting
and diagnosing errors during the syndrome extraction process in the
Steane 7-qubit quantum error correction code.
"""

from itertools import cycle
from typing import Any

from pecos.slr import Block, Comment, CReg, QReg
from pecos.slr.qeclib.generic.check_1flag import Check1Flag


def poly2qubits(poly: list[Any], data: QReg) -> list[Any]:
    """Convert polygon node IDs to qubit references.

    Args:
        poly: Polygon representation with node IDs and color.
        data: Quantum register containing the data qubits.

    Returns:
        List of qubit references corresponding to the polygon nodes.
    """
    return [data[q] for q in poly]


class SynExtractFlagged(Block):
    """Flagged syndrome extraction for Steane code with flag qubits for error detection."""

    def __init__(
        self,
        data: QReg,
        ancillas: QReg,
        checks: list,
        syn_x: CReg,
        syn_z: CReg,
        flag_bits_x: CReg,
        flag_bits_z: CReg,
    ) -> None:
        """Initialize flagged syndrome extraction.

        Args:
            data: Data qubit register.
            ancillas: Ancilla qubit register.
            checks: List of check operators to apply.
            syn_x: Classical register for X syndrome storage.
            syn_z: Classical register for Z syndrome storage.
            flag_bits_x: Classical register for X flag bit storage.
            flag_bits_z: Classical register for Z flag bit storage.

        Raises:
            ValueError: If register lengths don't match expected sizes.
        """
        if not (len(syn_x) == len(syn_z) == len(flag_bits_x) == len(flag_bits_z) == len(checks) == 3):
            msg = (
                f"Expected syndrome and flag registers of length 3 ({len(checks)} checks), "
                f"got syn_x={len(syn_x)}, syn_z={len(syn_z)}, "
                f"flag_bits_x={len(flag_bits_x)}, flag_bits_z={len(flag_bits_z)}"
            )
            raise ValueError(msg)
        a = cycle(range(len(ancillas)))
        sx = iter(range(len(syn_x)))
        sz = iter(range(len(syn_z)))
        fbx = iter(range(len(flag_bits_x)))
        fbz = iter(range(len(flag_bits_z)))

        super().__init__()

        pauli = "Z"
        for c in checks:
            data_ids = c[:-1]
            syn_id = next(sz)
            anc_id = next(a)
            flag_qubit_id = next(a)
            flag_bit_id = next(fbz)
            self.extend(
                Comment(f"Check['{pauli}', {data_ids}] -> {syn_z}[{syn_id}]"),
                Check1Flag(
                    d=poly2qubits(c, data),
                    ops=pauli,
                    a=ancillas[anc_id],
                    flag=ancillas[flag_qubit_id],
                    out=syn_z[syn_id],
                    out_flag=flag_bits_z[flag_bit_id],
                    with_barriers=False,
                ),
            )

        pauli = "X"
        for c in checks:
            data_ids = c[:-1]
            syn_id = next(sx)
            anc_id = next(a)
            flag_qubit_id = next(a)
            flag_bit_id = next(fbx)
            self.extend(
                Comment(f"Check['{pauli}', {data_ids}] -> {syn_x}[{syn_id}]"),
                Check1Flag(
                    d=poly2qubits(c, data),
                    ops=pauli,
                    a=ancillas[anc_id],
                    flag=ancillas[flag_qubit_id],
                    out=syn_x[syn_id],
                    out_flag=flag_bits_x[flag_bit_id],
                    with_barriers=False,
                ),
            )



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""QuEST state vector simulator for PECOS.

This module provides a quantum state vector simulator powered by the QuEST quantum simulation library,
enabling efficient simulation of arbitrary quantum circuits with full quantum state representation.
"""

from pecos.simulators.quest_statevec.state import QuestStateVec

__all__ = ["QuestStateVec"]



"""Test partial array consumption in SLR."""

from pecos.slr import Block, CReg, Main, QReg, SlrConverter
from pecos.slr.qeclib import qubit
from pecos.slr.qeclib.qubit.measures import Measure


class MeasureAncillas(Block):
    """Block for measuring ancilla qubits."""

    def __init__(self, data: QReg, ancilla: QReg, syndrome: CReg) -> None:
        """Initialize measurement block.

        Args:
            data: Data qubit register
            ancilla: Ancilla qubit register
            syndrome: Syndrome measurement register
        """
        super().__init__()
        self.data = data
        self.ancilla = ancilla
        self.syndrome = syndrome
        self.ops = [
            qubit.CX(data[0], ancilla[0]),
            Measure(ancilla) > syndrome,
        ]


prog = Main(
    data := QReg("data", 2),
    ancilla := QReg("ancilla", 1),
    syndrome := CReg("syndrome", 1),
    result := CReg("result", 2),
    MeasureAncillas(data, ancilla, syndrome),
    qubit.H(data[0]),
    Measure(data) > result,
)

print("Generated Guppy code:")
print("=" * 50)
print(SlrConverter(prog).guppy())
